"""Auto-generated :class:`zea.Operation` for all unary :mod:`keras.ops`
and :mod:`keras.ops.image` functions.

They can be used in zea pipelines like any other :class:`zea.Operation`, for example:

.. doctest::

    >>> from zea.ops.keras_ops import Squeeze

    >>> op = Squeeze(axis=1)

This file is generated automatically. Do not edit manually.
Generated with Keras 3.12.0
"""

import keras

from zea.internal.registry import ops_registry
from zea.ops.base import Lambda

class MissingKerasOps(ValueError):
    def __init__(self, class_name: str, func: str):
        super().__init__(
            f"Failed to create {class_name} with {func}. " +
            "This may be due to an incompatible version of `keras`. " +
            "Please try to upgrade `keras` to the latest version by running " +
            "`pip install --upgrade keras`."
        )


@ops_registry("keras.ops.abs")
class Abs(Lambda):
    """Operation wrapping keras.ops.abs."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.abs, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Abs", "keras.ops.abs") from e

@ops_registry("keras.ops.absolute")
class Absolute(Lambda):
    """Operation wrapping keras.ops.absolute."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.absolute, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Absolute", "keras.ops.absolute") from e

@ops_registry("keras.ops.all")
class All(Lambda):
    """Operation wrapping keras.ops.all."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.all, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("All", "keras.ops.all") from e

@ops_registry("keras.ops.amax")
class Amax(Lambda):
    """Operation wrapping keras.ops.amax."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.amax, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Amax", "keras.ops.amax") from e

@ops_registry("keras.ops.amin")
class Amin(Lambda):
    """Operation wrapping keras.ops.amin."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.amin, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Amin", "keras.ops.amin") from e

@ops_registry("keras.ops.angle")
class Angle(Lambda):
    """Operation wrapping keras.ops.angle."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.angle, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Angle", "keras.ops.angle") from e

@ops_registry("keras.ops.any")
class Any(Lambda):
    """Operation wrapping keras.ops.any."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.any, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Any", "keras.ops.any") from e

@ops_registry("keras.ops.arccos")
class Arccos(Lambda):
    """Operation wrapping keras.ops.arccos."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.arccos, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Arccos", "keras.ops.arccos") from e

@ops_registry("keras.ops.arccosh")
class Arccosh(Lambda):
    """Operation wrapping keras.ops.arccosh."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.arccosh, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Arccosh", "keras.ops.arccosh") from e

@ops_registry("keras.ops.arcsin")
class Arcsin(Lambda):
    """Operation wrapping keras.ops.arcsin."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.arcsin, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Arcsin", "keras.ops.arcsin") from e

@ops_registry("keras.ops.arcsinh")
class Arcsinh(Lambda):
    """Operation wrapping keras.ops.arcsinh."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.arcsinh, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Arcsinh", "keras.ops.arcsinh") from e

@ops_registry("keras.ops.arctan")
class Arctan(Lambda):
    """Operation wrapping keras.ops.arctan."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.arctan, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Arctan", "keras.ops.arctan") from e

@ops_registry("keras.ops.arctanh")
class Arctanh(Lambda):
    """Operation wrapping keras.ops.arctanh."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.arctanh, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Arctanh", "keras.ops.arctanh") from e

@ops_registry("keras.ops.argmax")
class Argmax(Lambda):
    """Operation wrapping keras.ops.argmax."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.argmax, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Argmax", "keras.ops.argmax") from e

@ops_registry("keras.ops.argmin")
class Argmin(Lambda):
    """Operation wrapping keras.ops.argmin."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.argmin, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Argmin", "keras.ops.argmin") from e

@ops_registry("keras.ops.argpartition")
class Argpartition(Lambda):
    """Operation wrapping keras.ops.argpartition."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.argpartition, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Argpartition", "keras.ops.argpartition") from e

@ops_registry("keras.ops.argsort")
class Argsort(Lambda):
    """Operation wrapping keras.ops.argsort."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.argsort, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Argsort", "keras.ops.argsort") from e

@ops_registry("keras.ops.array")
class Array(Lambda):
    """Operation wrapping keras.ops.array."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.array, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Array", "keras.ops.array") from e

@ops_registry("keras.ops.average")
class Average(Lambda):
    """Operation wrapping keras.ops.average."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.average, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Average", "keras.ops.average") from e

@ops_registry("keras.ops.bartlett")
class Bartlett(Lambda):
    """Operation wrapping keras.ops.bartlett."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.bartlett, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Bartlett", "keras.ops.bartlett") from e

@ops_registry("keras.ops.batch_normalization")
class BatchNormalization(Lambda):
    """Operation wrapping keras.ops.batch_normalization."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.batch_normalization, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("BatchNormalization", "keras.ops.batch_normalization") from e

@ops_registry("keras.ops.bincount")
class Bincount(Lambda):
    """Operation wrapping keras.ops.bincount."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.bincount, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Bincount", "keras.ops.bincount") from e

@ops_registry("keras.ops.bitwise_and")
class BitwiseAnd(Lambda):
    """Operation wrapping keras.ops.bitwise_and."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.bitwise_and, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("BitwiseAnd", "keras.ops.bitwise_and") from e

@ops_registry("keras.ops.bitwise_invert")
class BitwiseInvert(Lambda):
    """Operation wrapping keras.ops.bitwise_invert."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.bitwise_invert, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("BitwiseInvert", "keras.ops.bitwise_invert") from e

@ops_registry("keras.ops.bitwise_left_shift")
class BitwiseLeftShift(Lambda):
    """Operation wrapping keras.ops.bitwise_left_shift."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.bitwise_left_shift, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("BitwiseLeftShift", "keras.ops.bitwise_left_shift") from e

@ops_registry("keras.ops.bitwise_not")
class BitwiseNot(Lambda):
    """Operation wrapping keras.ops.bitwise_not."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.bitwise_not, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("BitwiseNot", "keras.ops.bitwise_not") from e

@ops_registry("keras.ops.bitwise_or")
class BitwiseOr(Lambda):
    """Operation wrapping keras.ops.bitwise_or."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.bitwise_or, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("BitwiseOr", "keras.ops.bitwise_or") from e

@ops_registry("keras.ops.bitwise_right_shift")
class BitwiseRightShift(Lambda):
    """Operation wrapping keras.ops.bitwise_right_shift."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.bitwise_right_shift, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("BitwiseRightShift", "keras.ops.bitwise_right_shift") from e

@ops_registry("keras.ops.bitwise_xor")
class BitwiseXor(Lambda):
    """Operation wrapping keras.ops.bitwise_xor."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.bitwise_xor, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("BitwiseXor", "keras.ops.bitwise_xor") from e

@ops_registry("keras.ops.blackman")
class Blackman(Lambda):
    """Operation wrapping keras.ops.blackman."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.blackman, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Blackman", "keras.ops.blackman") from e

@ops_registry("keras.ops.broadcast_to")
class BroadcastTo(Lambda):
    """Operation wrapping keras.ops.broadcast_to."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.broadcast_to, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("BroadcastTo", "keras.ops.broadcast_to") from e

@ops_registry("keras.ops.cast")
class Cast(Lambda):
    """Operation wrapping keras.ops.cast."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.cast, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Cast", "keras.ops.cast") from e

@ops_registry("keras.ops.cbrt")
class Cbrt(Lambda):
    """Operation wrapping keras.ops.cbrt."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.cbrt, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Cbrt", "keras.ops.cbrt") from e

@ops_registry("keras.ops.ceil")
class Ceil(Lambda):
    """Operation wrapping keras.ops.ceil."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.ceil, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Ceil", "keras.ops.ceil") from e

@ops_registry("keras.ops.celu")
class Celu(Lambda):
    """Operation wrapping keras.ops.celu."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.celu, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Celu", "keras.ops.celu") from e

@ops_registry("keras.ops.cholesky")
class Cholesky(Lambda):
    """Operation wrapping keras.ops.cholesky."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.cholesky, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Cholesky", "keras.ops.cholesky") from e

@ops_registry("keras.ops.cholesky_inverse")
class CholeskyInverse(Lambda):
    """Operation wrapping keras.ops.cholesky_inverse."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.cholesky_inverse, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("CholeskyInverse", "keras.ops.cholesky_inverse") from e

@ops_registry("keras.ops.clip")
class Clip(Lambda):
    """Operation wrapping keras.ops.clip."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.clip, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Clip", "keras.ops.clip") from e

@ops_registry("keras.ops.conj")
class Conj(Lambda):
    """Operation wrapping keras.ops.conj."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.conj, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Conj", "keras.ops.conj") from e

@ops_registry("keras.ops.conjugate")
class Conjugate(Lambda):
    """Operation wrapping keras.ops.conjugate."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.conjugate, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Conjugate", "keras.ops.conjugate") from e

@ops_registry("keras.ops.convert_to_numpy")
class ConvertToNumpy(Lambda):
    """Operation wrapping keras.ops.convert_to_numpy."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.convert_to_numpy, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("ConvertToNumpy", "keras.ops.convert_to_numpy") from e

@ops_registry("keras.ops.convert_to_tensor")
class ConvertToTensor(Lambda):
    """Operation wrapping keras.ops.convert_to_tensor."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.convert_to_tensor, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("ConvertToTensor", "keras.ops.convert_to_tensor") from e

@ops_registry("keras.ops.copy")
class Copy(Lambda):
    """Operation wrapping keras.ops.copy."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.copy, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Copy", "keras.ops.copy") from e

@ops_registry("keras.ops.corrcoef")
class Corrcoef(Lambda):
    """Operation wrapping keras.ops.corrcoef."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.corrcoef, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Corrcoef", "keras.ops.corrcoef") from e

@ops_registry("keras.ops.cos")
class Cos(Lambda):
    """Operation wrapping keras.ops.cos."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.cos, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Cos", "keras.ops.cos") from e

@ops_registry("keras.ops.cosh")
class Cosh(Lambda):
    """Operation wrapping keras.ops.cosh."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.cosh, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Cosh", "keras.ops.cosh") from e

@ops_registry("keras.ops.count_nonzero")
class CountNonzero(Lambda):
    """Operation wrapping keras.ops.count_nonzero."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.count_nonzero, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("CountNonzero", "keras.ops.count_nonzero") from e

@ops_registry("keras.ops.cumprod")
class Cumprod(Lambda):
    """Operation wrapping keras.ops.cumprod."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.cumprod, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Cumprod", "keras.ops.cumprod") from e

@ops_registry("keras.ops.cumsum")
class Cumsum(Lambda):
    """Operation wrapping keras.ops.cumsum."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.cumsum, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Cumsum", "keras.ops.cumsum") from e

@ops_registry("keras.ops.deg2rad")
class Deg2rad(Lambda):
    """Operation wrapping keras.ops.deg2rad."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.deg2rad, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Deg2rad", "keras.ops.deg2rad") from e

@ops_registry("keras.ops.det")
class Det(Lambda):
    """Operation wrapping keras.ops.det."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.det, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Det", "keras.ops.det") from e

@ops_registry("keras.ops.diag")
class Diag(Lambda):
    """Operation wrapping keras.ops.diag."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.diag, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Diag", "keras.ops.diag") from e

@ops_registry("keras.ops.diagflat")
class Diagflat(Lambda):
    """Operation wrapping keras.ops.diagflat."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.diagflat, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Diagflat", "keras.ops.diagflat") from e

@ops_registry("keras.ops.diagonal")
class Diagonal(Lambda):
    """Operation wrapping keras.ops.diagonal."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.diagonal, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Diagonal", "keras.ops.diagonal") from e

@ops_registry("keras.ops.digitize")
class Digitize(Lambda):
    """Operation wrapping keras.ops.digitize."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.digitize, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Digitize", "keras.ops.digitize") from e

@ops_registry("keras.ops.dtype")
class Dtype(Lambda):
    """Operation wrapping keras.ops.dtype."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.dtype, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Dtype", "keras.ops.dtype") from e

@ops_registry("keras.ops.eig")
class Eig(Lambda):
    """Operation wrapping keras.ops.eig."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.eig, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Eig", "keras.ops.eig") from e

@ops_registry("keras.ops.eigh")
class Eigh(Lambda):
    """Operation wrapping keras.ops.eigh."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.eigh, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Eigh", "keras.ops.eigh") from e

@ops_registry("keras.ops.elu")
class Elu(Lambda):
    """Operation wrapping keras.ops.elu."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.elu, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Elu", "keras.ops.elu") from e

@ops_registry("keras.ops.erf")
class Erf(Lambda):
    """Operation wrapping keras.ops.erf."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.erf, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Erf", "keras.ops.erf") from e

@ops_registry("keras.ops.erfinv")
class Erfinv(Lambda):
    """Operation wrapping keras.ops.erfinv."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.erfinv, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Erfinv", "keras.ops.erfinv") from e

@ops_registry("keras.ops.exp")
class Exp(Lambda):
    """Operation wrapping keras.ops.exp."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.exp, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Exp", "keras.ops.exp") from e

@ops_registry("keras.ops.exp2")
class Exp2(Lambda):
    """Operation wrapping keras.ops.exp2."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.exp2, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Exp2", "keras.ops.exp2") from e

@ops_registry("keras.ops.expand_dims")
class ExpandDims(Lambda):
    """Operation wrapping keras.ops.expand_dims."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.expand_dims, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("ExpandDims", "keras.ops.expand_dims") from e

@ops_registry("keras.ops.expm1")
class Expm1(Lambda):
    """Operation wrapping keras.ops.expm1."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.expm1, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Expm1", "keras.ops.expm1") from e

@ops_registry("keras.ops.extract_sequences")
class ExtractSequences(Lambda):
    """Operation wrapping keras.ops.extract_sequences."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.extract_sequences, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("ExtractSequences", "keras.ops.extract_sequences") from e

@ops_registry("keras.ops.fft")
class Fft(Lambda):
    """Operation wrapping keras.ops.fft."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.fft, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Fft", "keras.ops.fft") from e

@ops_registry("keras.ops.fft2")
class Fft2(Lambda):
    """Operation wrapping keras.ops.fft2."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.fft2, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Fft2", "keras.ops.fft2") from e

@ops_registry("keras.ops.flip")
class Flip(Lambda):
    """Operation wrapping keras.ops.flip."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.flip, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Flip", "keras.ops.flip") from e

@ops_registry("keras.ops.floor")
class Floor(Lambda):
    """Operation wrapping keras.ops.floor."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.floor, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Floor", "keras.ops.floor") from e

@ops_registry("keras.ops.full_like")
class FullLike(Lambda):
    """Operation wrapping keras.ops.full_like."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.full_like, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("FullLike", "keras.ops.full_like") from e

@ops_registry("keras.ops.gelu")
class Gelu(Lambda):
    """Operation wrapping keras.ops.gelu."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.gelu, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Gelu", "keras.ops.gelu") from e

@ops_registry("keras.ops.get_item")
class GetItem(Lambda):
    """Operation wrapping keras.ops.get_item."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.get_item, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("GetItem", "keras.ops.get_item") from e

@ops_registry("keras.ops.glu")
class Glu(Lambda):
    """Operation wrapping keras.ops.glu."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.glu, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Glu", "keras.ops.glu") from e

@ops_registry("keras.ops.hamming")
class Hamming(Lambda):
    """Operation wrapping keras.ops.hamming."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.hamming, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Hamming", "keras.ops.hamming") from e

@ops_registry("keras.ops.hanning")
class Hanning(Lambda):
    """Operation wrapping keras.ops.hanning."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.hanning, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Hanning", "keras.ops.hanning") from e

@ops_registry("keras.ops.hard_shrink")
class HardShrink(Lambda):
    """Operation wrapping keras.ops.hard_shrink."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.hard_shrink, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("HardShrink", "keras.ops.hard_shrink") from e

@ops_registry("keras.ops.hard_sigmoid")
class HardSigmoid(Lambda):
    """Operation wrapping keras.ops.hard_sigmoid."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.hard_sigmoid, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("HardSigmoid", "keras.ops.hard_sigmoid") from e

@ops_registry("keras.ops.hard_silu")
class HardSilu(Lambda):
    """Operation wrapping keras.ops.hard_silu."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.hard_silu, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("HardSilu", "keras.ops.hard_silu") from e

@ops_registry("keras.ops.hard_swish")
class HardSwish(Lambda):
    """Operation wrapping keras.ops.hard_swish."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.hard_swish, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("HardSwish", "keras.ops.hard_swish") from e

@ops_registry("keras.ops.hard_tanh")
class HardTanh(Lambda):
    """Operation wrapping keras.ops.hard_tanh."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.hard_tanh, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("HardTanh", "keras.ops.hard_tanh") from e

@ops_registry("keras.ops.histogram")
class Histogram(Lambda):
    """Operation wrapping keras.ops.histogram."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.histogram, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Histogram", "keras.ops.histogram") from e

@ops_registry("keras.ops.ifft2")
class Ifft2(Lambda):
    """Operation wrapping keras.ops.ifft2."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.ifft2, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Ifft2", "keras.ops.ifft2") from e

@ops_registry("keras.ops.imag")
class Imag(Lambda):
    """Operation wrapping keras.ops.imag."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.imag, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Imag", "keras.ops.imag") from e

@ops_registry("keras.ops.inv")
class Inv(Lambda):
    """Operation wrapping keras.ops.inv."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.inv, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Inv", "keras.ops.inv") from e

@ops_registry("keras.ops.irfft")
class Irfft(Lambda):
    """Operation wrapping keras.ops.irfft."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.irfft, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Irfft", "keras.ops.irfft") from e

@ops_registry("keras.ops.is_tensor")
class IsTensor(Lambda):
    """Operation wrapping keras.ops.is_tensor."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.is_tensor, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("IsTensor", "keras.ops.is_tensor") from e

@ops_registry("keras.ops.isfinite")
class Isfinite(Lambda):
    """Operation wrapping keras.ops.isfinite."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.isfinite, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Isfinite", "keras.ops.isfinite") from e

@ops_registry("keras.ops.isinf")
class Isinf(Lambda):
    """Operation wrapping keras.ops.isinf."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.isinf, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Isinf", "keras.ops.isinf") from e

@ops_registry("keras.ops.isnan")
class Isnan(Lambda):
    """Operation wrapping keras.ops.isnan."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.isnan, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Isnan", "keras.ops.isnan") from e

@ops_registry("keras.ops.isneginf")
class Isneginf(Lambda):
    """Operation wrapping keras.ops.isneginf."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.isneginf, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Isneginf", "keras.ops.isneginf") from e

@ops_registry("keras.ops.isposinf")
class Isposinf(Lambda):
    """Operation wrapping keras.ops.isposinf."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.isposinf, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Isposinf", "keras.ops.isposinf") from e

@ops_registry("keras.ops.isreal")
class Isreal(Lambda):
    """Operation wrapping keras.ops.isreal."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.isreal, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Isreal", "keras.ops.isreal") from e

@ops_registry("keras.ops.istft")
class Istft(Lambda):
    """Operation wrapping keras.ops.istft."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.istft, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Istft", "keras.ops.istft") from e

@ops_registry("keras.ops.kaiser")
class Kaiser(Lambda):
    """Operation wrapping keras.ops.kaiser."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.kaiser, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Kaiser", "keras.ops.kaiser") from e

@ops_registry("keras.ops.layer_normalization")
class LayerNormalization(Lambda):
    """Operation wrapping keras.ops.layer_normalization."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.layer_normalization, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("LayerNormalization", "keras.ops.layer_normalization") from e

@ops_registry("keras.ops.leaky_relu")
class LeakyRelu(Lambda):
    """Operation wrapping keras.ops.leaky_relu."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.leaky_relu, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("LeakyRelu", "keras.ops.leaky_relu") from e

@ops_registry("keras.ops.left_shift")
class LeftShift(Lambda):
    """Operation wrapping keras.ops.left_shift."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.left_shift, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("LeftShift", "keras.ops.left_shift") from e

@ops_registry("keras.ops.log")
class Log(Lambda):
    """Operation wrapping keras.ops.log."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.log, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Log", "keras.ops.log") from e

@ops_registry("keras.ops.log10")
class Log10(Lambda):
    """Operation wrapping keras.ops.log10."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.log10, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Log10", "keras.ops.log10") from e

@ops_registry("keras.ops.log1p")
class Log1p(Lambda):
    """Operation wrapping keras.ops.log1p."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.log1p, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Log1p", "keras.ops.log1p") from e

@ops_registry("keras.ops.log2")
class Log2(Lambda):
    """Operation wrapping keras.ops.log2."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.log2, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Log2", "keras.ops.log2") from e

@ops_registry("keras.ops.log_sigmoid")
class LogSigmoid(Lambda):
    """Operation wrapping keras.ops.log_sigmoid."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.log_sigmoid, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("LogSigmoid", "keras.ops.log_sigmoid") from e

@ops_registry("keras.ops.log_softmax")
class LogSoftmax(Lambda):
    """Operation wrapping keras.ops.log_softmax."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.log_softmax, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("LogSoftmax", "keras.ops.log_softmax") from e

@ops_registry("keras.ops.logdet")
class Logdet(Lambda):
    """Operation wrapping keras.ops.logdet."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.logdet, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Logdet", "keras.ops.logdet") from e

@ops_registry("keras.ops.logical_not")
class LogicalNot(Lambda):
    """Operation wrapping keras.ops.logical_not."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.logical_not, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("LogicalNot", "keras.ops.logical_not") from e

@ops_registry("keras.ops.logsumexp")
class Logsumexp(Lambda):
    """Operation wrapping keras.ops.logsumexp."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.logsumexp, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Logsumexp", "keras.ops.logsumexp") from e

@ops_registry("keras.ops.lu_factor")
class LuFactor(Lambda):
    """Operation wrapping keras.ops.lu_factor."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.lu_factor, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("LuFactor", "keras.ops.lu_factor") from e

@ops_registry("keras.ops.max")
class Max(Lambda):
    """Operation wrapping keras.ops.max."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.max, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Max", "keras.ops.max") from e

@ops_registry("keras.ops.mean")
class Mean(Lambda):
    """Operation wrapping keras.ops.mean."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.mean, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Mean", "keras.ops.mean") from e

@ops_registry("keras.ops.median")
class Median(Lambda):
    """Operation wrapping keras.ops.median."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.median, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Median", "keras.ops.median") from e

@ops_registry("keras.ops.meshgrid")
class Meshgrid(Lambda):
    """Operation wrapping keras.ops.meshgrid."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.meshgrid, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Meshgrid", "keras.ops.meshgrid") from e

@ops_registry("keras.ops.min")
class Min(Lambda):
    """Operation wrapping keras.ops.min."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.min, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Min", "keras.ops.min") from e

@ops_registry("keras.ops.moments")
class Moments(Lambda):
    """Operation wrapping keras.ops.moments."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.moments, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Moments", "keras.ops.moments") from e

@ops_registry("keras.ops.moveaxis")
class Moveaxis(Lambda):
    """Operation wrapping keras.ops.moveaxis."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.moveaxis, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Moveaxis", "keras.ops.moveaxis") from e

@ops_registry("keras.ops.nan_to_num")
class NanToNum(Lambda):
    """Operation wrapping keras.ops.nan_to_num."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.nan_to_num, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("NanToNum", "keras.ops.nan_to_num") from e

@ops_registry("keras.ops.ndim")
class Ndim(Lambda):
    """Operation wrapping keras.ops.ndim."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.ndim, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Ndim", "keras.ops.ndim") from e

@ops_registry("keras.ops.negative")
class Negative(Lambda):
    """Operation wrapping keras.ops.negative."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.negative, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Negative", "keras.ops.negative") from e

@ops_registry("keras.ops.nonzero")
class Nonzero(Lambda):
    """Operation wrapping keras.ops.nonzero."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.nonzero, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Nonzero", "keras.ops.nonzero") from e

@ops_registry("keras.ops.norm")
class Norm(Lambda):
    """Operation wrapping keras.ops.norm."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.norm, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Norm", "keras.ops.norm") from e

@ops_registry("keras.ops.normalize")
class Normalize(Lambda):
    """Operation wrapping keras.ops.normalize."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.normalize, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Normalize", "keras.ops.normalize") from e

@ops_registry("keras.ops.one_hot")
class OneHot(Lambda):
    """Operation wrapping keras.ops.one_hot."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.one_hot, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("OneHot", "keras.ops.one_hot") from e

@ops_registry("keras.ops.ones_like")
class OnesLike(Lambda):
    """Operation wrapping keras.ops.ones_like."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.ones_like, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("OnesLike", "keras.ops.ones_like") from e

@ops_registry("keras.ops.pad")
class Pad(Lambda):
    """Operation wrapping keras.ops.pad."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.pad, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Pad", "keras.ops.pad") from e

@ops_registry("keras.ops.prod")
class Prod(Lambda):
    """Operation wrapping keras.ops.prod."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.prod, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Prod", "keras.ops.prod") from e

@ops_registry("keras.ops.qr")
class Qr(Lambda):
    """Operation wrapping keras.ops.qr."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.qr, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Qr", "keras.ops.qr") from e

@ops_registry("keras.ops.quantile")
class Quantile(Lambda):
    """Operation wrapping keras.ops.quantile."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.quantile, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Quantile", "keras.ops.quantile") from e

@ops_registry("keras.ops.ravel")
class Ravel(Lambda):
    """Operation wrapping keras.ops.ravel."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.ravel, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Ravel", "keras.ops.ravel") from e

@ops_registry("keras.ops.real")
class Real(Lambda):
    """Operation wrapping keras.ops.real."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.real, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Real", "keras.ops.real") from e

@ops_registry("keras.ops.reciprocal")
class Reciprocal(Lambda):
    """Operation wrapping keras.ops.reciprocal."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.reciprocal, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Reciprocal", "keras.ops.reciprocal") from e

@ops_registry("keras.ops.relu")
class Relu(Lambda):
    """Operation wrapping keras.ops.relu."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.relu, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Relu", "keras.ops.relu") from e

@ops_registry("keras.ops.relu6")
class Relu6(Lambda):
    """Operation wrapping keras.ops.relu6."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.relu6, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Relu6", "keras.ops.relu6") from e

@ops_registry("keras.ops.repeat")
class Repeat(Lambda):
    """Operation wrapping keras.ops.repeat."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.repeat, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Repeat", "keras.ops.repeat") from e

@ops_registry("keras.ops.reshape")
class Reshape(Lambda):
    """Operation wrapping keras.ops.reshape."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.reshape, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Reshape", "keras.ops.reshape") from e

@ops_registry("keras.ops.rfft")
class Rfft(Lambda):
    """Operation wrapping keras.ops.rfft."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.rfft, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Rfft", "keras.ops.rfft") from e

@ops_registry("keras.ops.right_shift")
class RightShift(Lambda):
    """Operation wrapping keras.ops.right_shift."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.right_shift, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("RightShift", "keras.ops.right_shift") from e

@ops_registry("keras.ops.rms_normalization")
class RmsNormalization(Lambda):
    """Operation wrapping keras.ops.rms_normalization."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.rms_normalization, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("RmsNormalization", "keras.ops.rms_normalization") from e

@ops_registry("keras.ops.roll")
class Roll(Lambda):
    """Operation wrapping keras.ops.roll."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.roll, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Roll", "keras.ops.roll") from e

@ops_registry("keras.ops.round")
class Round(Lambda):
    """Operation wrapping keras.ops.round."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.round, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Round", "keras.ops.round") from e

@ops_registry("keras.ops.rsqrt")
class Rsqrt(Lambda):
    """Operation wrapping keras.ops.rsqrt."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.rsqrt, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Rsqrt", "keras.ops.rsqrt") from e

@ops_registry("keras.ops.saturate_cast")
class SaturateCast(Lambda):
    """Operation wrapping keras.ops.saturate_cast."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.saturate_cast, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("SaturateCast", "keras.ops.saturate_cast") from e

@ops_registry("keras.ops.selu")
class Selu(Lambda):
    """Operation wrapping keras.ops.selu."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.selu, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Selu", "keras.ops.selu") from e

@ops_registry("keras.ops.shape")
class Shape(Lambda):
    """Operation wrapping keras.ops.shape."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.shape, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Shape", "keras.ops.shape") from e

@ops_registry("keras.ops.sigmoid")
class Sigmoid(Lambda):
    """Operation wrapping keras.ops.sigmoid."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.sigmoid, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Sigmoid", "keras.ops.sigmoid") from e

@ops_registry("keras.ops.sign")
class Sign(Lambda):
    """Operation wrapping keras.ops.sign."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.sign, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Sign", "keras.ops.sign") from e

@ops_registry("keras.ops.signbit")
class Signbit(Lambda):
    """Operation wrapping keras.ops.signbit."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.signbit, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Signbit", "keras.ops.signbit") from e

@ops_registry("keras.ops.silu")
class Silu(Lambda):
    """Operation wrapping keras.ops.silu."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.silu, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Silu", "keras.ops.silu") from e

@ops_registry("keras.ops.sin")
class Sin(Lambda):
    """Operation wrapping keras.ops.sin."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.sin, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Sin", "keras.ops.sin") from e

@ops_registry("keras.ops.sinh")
class Sinh(Lambda):
    """Operation wrapping keras.ops.sinh."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.sinh, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Sinh", "keras.ops.sinh") from e

@ops_registry("keras.ops.size")
class Size(Lambda):
    """Operation wrapping keras.ops.size."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.size, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Size", "keras.ops.size") from e

@ops_registry("keras.ops.slogdet")
class Slogdet(Lambda):
    """Operation wrapping keras.ops.slogdet."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.slogdet, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Slogdet", "keras.ops.slogdet") from e

@ops_registry("keras.ops.soft_shrink")
class SoftShrink(Lambda):
    """Operation wrapping keras.ops.soft_shrink."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.soft_shrink, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("SoftShrink", "keras.ops.soft_shrink") from e

@ops_registry("keras.ops.softmax")
class Softmax(Lambda):
    """Operation wrapping keras.ops.softmax."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.softmax, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Softmax", "keras.ops.softmax") from e

@ops_registry("keras.ops.softplus")
class Softplus(Lambda):
    """Operation wrapping keras.ops.softplus."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.softplus, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Softplus", "keras.ops.softplus") from e

@ops_registry("keras.ops.softsign")
class Softsign(Lambda):
    """Operation wrapping keras.ops.softsign."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.softsign, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Softsign", "keras.ops.softsign") from e

@ops_registry("keras.ops.sort")
class Sort(Lambda):
    """Operation wrapping keras.ops.sort."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.sort, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Sort", "keras.ops.sort") from e

@ops_registry("keras.ops.sparse_plus")
class SparsePlus(Lambda):
    """Operation wrapping keras.ops.sparse_plus."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.sparse_plus, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("SparsePlus", "keras.ops.sparse_plus") from e

@ops_registry("keras.ops.sparse_sigmoid")
class SparseSigmoid(Lambda):
    """Operation wrapping keras.ops.sparse_sigmoid."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.sparse_sigmoid, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("SparseSigmoid", "keras.ops.sparse_sigmoid") from e

@ops_registry("keras.ops.sparsemax")
class Sparsemax(Lambda):
    """Operation wrapping keras.ops.sparsemax."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.sparsemax, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Sparsemax", "keras.ops.sparsemax") from e

@ops_registry("keras.ops.split")
class Split(Lambda):
    """Operation wrapping keras.ops.split."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.split, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Split", "keras.ops.split") from e

@ops_registry("keras.ops.sqrt")
class Sqrt(Lambda):
    """Operation wrapping keras.ops.sqrt."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.sqrt, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Sqrt", "keras.ops.sqrt") from e

@ops_registry("keras.ops.square")
class Square(Lambda):
    """Operation wrapping keras.ops.square."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.square, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Square", "keras.ops.square") from e

@ops_registry("keras.ops.squareplus")
class Squareplus(Lambda):
    """Operation wrapping keras.ops.squareplus."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.squareplus, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Squareplus", "keras.ops.squareplus") from e

@ops_registry("keras.ops.squeeze")
class Squeeze(Lambda):
    """Operation wrapping keras.ops.squeeze."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.squeeze, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Squeeze", "keras.ops.squeeze") from e

@ops_registry("keras.ops.stack")
class Stack(Lambda):
    """Operation wrapping keras.ops.stack."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.stack, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Stack", "keras.ops.stack") from e

@ops_registry("keras.ops.std")
class Std(Lambda):
    """Operation wrapping keras.ops.std."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.std, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Std", "keras.ops.std") from e

@ops_registry("keras.ops.stft")
class Stft(Lambda):
    """Operation wrapping keras.ops.stft."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.stft, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Stft", "keras.ops.stft") from e

@ops_registry("keras.ops.sum")
class Sum(Lambda):
    """Operation wrapping keras.ops.sum."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.sum, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Sum", "keras.ops.sum") from e

@ops_registry("keras.ops.svd")
class Svd(Lambda):
    """Operation wrapping keras.ops.svd."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.svd, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Svd", "keras.ops.svd") from e

@ops_registry("keras.ops.swapaxes")
class Swapaxes(Lambda):
    """Operation wrapping keras.ops.swapaxes."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.swapaxes, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Swapaxes", "keras.ops.swapaxes") from e

@ops_registry("keras.ops.swish")
class Swish(Lambda):
    """Operation wrapping keras.ops.swish."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.swish, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Swish", "keras.ops.swish") from e

@ops_registry("keras.ops.take")
class Take(Lambda):
    """Operation wrapping keras.ops.take."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.take, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Take", "keras.ops.take") from e

@ops_registry("keras.ops.take_along_axis")
class TakeAlongAxis(Lambda):
    """Operation wrapping keras.ops.take_along_axis."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.take_along_axis, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("TakeAlongAxis", "keras.ops.take_along_axis") from e

@ops_registry("keras.ops.tan")
class Tan(Lambda):
    """Operation wrapping keras.ops.tan."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.tan, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Tan", "keras.ops.tan") from e

@ops_registry("keras.ops.tanh")
class Tanh(Lambda):
    """Operation wrapping keras.ops.tanh."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.tanh, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Tanh", "keras.ops.tanh") from e

@ops_registry("keras.ops.tanh_shrink")
class TanhShrink(Lambda):
    """Operation wrapping keras.ops.tanh_shrink."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.tanh_shrink, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("TanhShrink", "keras.ops.tanh_shrink") from e

@ops_registry("keras.ops.threshold")
class Threshold(Lambda):
    """Operation wrapping keras.ops.threshold."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.threshold, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Threshold", "keras.ops.threshold") from e

@ops_registry("keras.ops.tile")
class Tile(Lambda):
    """Operation wrapping keras.ops.tile."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.tile, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Tile", "keras.ops.tile") from e

@ops_registry("keras.ops.top_k")
class TopK(Lambda):
    """Operation wrapping keras.ops.top_k."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.top_k, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("TopK", "keras.ops.top_k") from e

@ops_registry("keras.ops.trace")
class Trace(Lambda):
    """Operation wrapping keras.ops.trace."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.trace, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Trace", "keras.ops.trace") from e

@ops_registry("keras.ops.transpose")
class Transpose(Lambda):
    """Operation wrapping keras.ops.transpose."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.transpose, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Transpose", "keras.ops.transpose") from e

@ops_registry("keras.ops.tril")
class Tril(Lambda):
    """Operation wrapping keras.ops.tril."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.tril, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Tril", "keras.ops.tril") from e

@ops_registry("keras.ops.triu")
class Triu(Lambda):
    """Operation wrapping keras.ops.triu."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.triu, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Triu", "keras.ops.triu") from e

@ops_registry("keras.ops.trunc")
class Trunc(Lambda):
    """Operation wrapping keras.ops.trunc."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.trunc, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Trunc", "keras.ops.trunc") from e

@ops_registry("keras.ops.unfold")
class Unfold(Lambda):
    """Operation wrapping keras.ops.unfold."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.unfold, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Unfold", "keras.ops.unfold") from e

@ops_registry("keras.ops.unstack")
class Unstack(Lambda):
    """Operation wrapping keras.ops.unstack."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.unstack, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Unstack", "keras.ops.unstack") from e

@ops_registry("keras.ops.var")
class Var(Lambda):
    """Operation wrapping keras.ops.var."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.var, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Var", "keras.ops.var") from e

@ops_registry("keras.ops.view")
class View(Lambda):
    """Operation wrapping keras.ops.view."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.view, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("View", "keras.ops.view") from e

@ops_registry("keras.ops.view_as_complex")
class ViewAsComplex(Lambda):
    """Operation wrapping keras.ops.view_as_complex."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.view_as_complex, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("ViewAsComplex", "keras.ops.view_as_complex") from e

@ops_registry("keras.ops.view_as_real")
class ViewAsReal(Lambda):
    """Operation wrapping keras.ops.view_as_real."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.view_as_real, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("ViewAsReal", "keras.ops.view_as_real") from e

@ops_registry("keras.ops.zeros_like")
class ZerosLike(Lambda):
    """Operation wrapping keras.ops.zeros_like."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.zeros_like, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("ZerosLike", "keras.ops.zeros_like") from e

@ops_registry("keras.ops.image.affine_transform")
class AffineTransform(Lambda):
    """Operation wrapping keras.ops.image.affine_transform."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.image.affine_transform, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("AffineTransform", "keras.ops.image.affine_transform") from e

@ops_registry("keras.ops.image.crop_images")
class CropImages(Lambda):
    """Operation wrapping keras.ops.image.crop_images."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.image.crop_images, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("CropImages", "keras.ops.image.crop_images") from e

@ops_registry("keras.ops.image.elastic_transform")
class ElasticTransform(Lambda):
    """Operation wrapping keras.ops.image.elastic_transform."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.image.elastic_transform, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("ElasticTransform", "keras.ops.image.elastic_transform") from e

@ops_registry("keras.ops.image.extract_patches")
class ExtractPatches(Lambda):
    """Operation wrapping keras.ops.image.extract_patches."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.image.extract_patches, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("ExtractPatches", "keras.ops.image.extract_patches") from e

@ops_registry("keras.ops.image.gaussian_blur")
class GaussianBlur(Lambda):
    """Operation wrapping keras.ops.image.gaussian_blur."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.image.gaussian_blur, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("GaussianBlur", "keras.ops.image.gaussian_blur") from e

@ops_registry("keras.ops.image.hsv_to_rgb")
class HsvToRgb(Lambda):
    """Operation wrapping keras.ops.image.hsv_to_rgb."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.image.hsv_to_rgb, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("HsvToRgb", "keras.ops.image.hsv_to_rgb") from e

@ops_registry("keras.ops.image.pad_images")
class PadImages(Lambda):
    """Operation wrapping keras.ops.image.pad_images."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.image.pad_images, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("PadImages", "keras.ops.image.pad_images") from e

@ops_registry("keras.ops.image.perspective_transform")
class PerspectiveTransform(Lambda):
    """Operation wrapping keras.ops.image.perspective_transform."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.image.perspective_transform, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("PerspectiveTransform", "keras.ops.image.perspective_transform") from e

@ops_registry("keras.ops.image.resize")
class Resize(Lambda):
    """Operation wrapping keras.ops.image.resize."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.image.resize, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("Resize", "keras.ops.image.resize") from e

@ops_registry("keras.ops.image.rgb_to_grayscale")
class RgbToGrayscale(Lambda):
    """Operation wrapping keras.ops.image.rgb_to_grayscale."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.image.rgb_to_grayscale, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("RgbToGrayscale", "keras.ops.image.rgb_to_grayscale") from e

@ops_registry("keras.ops.image.rgb_to_hsv")
class RgbToHsv(Lambda):
    """Operation wrapping keras.ops.image.rgb_to_hsv."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.image.rgb_to_hsv, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("RgbToHsv", "keras.ops.image.rgb_to_hsv") from e

@ops_registry("keras.ops.image.scale_and_translate")
class ScaleAndTranslate(Lambda):
    """Operation wrapping keras.ops.image.scale_and_translate."""

    def __init__(self, **kwargs):
        try:
            super().__init__(func=keras.ops.image.scale_and_translate, **kwargs)
        except AttributeError as e:
            raise MissingKerasOps("ScaleAndTranslate", "keras.ops.image.scale_and_translate") from e



"""Tensorflow Ultrasound Beamforming Library.

Initialize modules for registries.
"""

import sys
from pathlib import PosixPath

import numpy as np

# Convert PosixPath objects to strings in sys.path
# this is necessary due to weird TF bug when importing
sys.path = [str(p) if isinstance(p, PosixPath) else p for p in sys.path]

import tensorflow as tf  # noqa: E402

from .dataloader import make_dataloader  # noqa: E402



from typing import Tuple

import keras
import numpy as np
from keras import ops

from zea import log
from zea.beamform.beamformer import tof_correction
from zea.display import scan_convert
from zea.func.tensor import (
    apply_along_axis,
    correlate,
    extend_n_dims,
    gaussian_filter,
    reshape_axis,
)
from zea.func.ultrasound import (
    channels_to_complex,
    complex_to_channels,
    demodulate,
    envelope_detect,
    get_band_pass_filter,
    get_low_pass_iq_filter,
    log_compress,
    upmix,
)
from zea.internal.core import (
    DEFAULT_DYNAMIC_RANGE,
    DataTypes,
)
from zea.internal.registry import ops_registry
from zea.ops.base import Filter, Operation
from zea.simulator import simulate_rf
from zea.utils import canonicalize_axis


@ops_registry("simulate_rf")
class Simulate(Operation):
    """Simulate RF data."""

    # Define operation-specific static parameters
    STATIC_PARAMS = ["n_ax", "apply_lens_correction"]
    ADD_OUTPUT_KEYS = ["n_ch"]

    def __init__(self, **kwargs):
        super().__init__(
            output_data_type=DataTypes.RAW_DATA,
            **kwargs,
        )

    def call(
        self,
        scatterer_positions,
        scatterer_magnitudes,
        probe_geometry,
        apply_lens_correction,
        lens_thickness,
        lens_sound_speed,
        sound_speed,
        n_ax,
        center_frequency,
        sampling_frequency,
        t0_delays,
        initial_times,
        element_width,
        attenuation_coef,
        tx_apodizations,
        **kwargs,
    ):
        simulate_kwargs = {
            "probe_geometry": probe_geometry,
            "apply_lens_correction": apply_lens_correction,
            "lens_thickness": lens_thickness,
            "lens_sound_speed": lens_sound_speed,
            "sound_speed": sound_speed,
            "n_ax": n_ax,
            "center_frequency": center_frequency,
            "sampling_frequency": sampling_frequency,
            "t0_delays": t0_delays,
            "initial_times": initial_times,
            "element_width": element_width,
            "attenuation_coef": attenuation_coef,
            "tx_apodizations": tx_apodizations,
        }
        if not self.with_batch_dim:
            simulated_rf = simulate_rf(
                scatterer_positions=scatterer_positions,
                scatterer_magnitudes=scatterer_magnitudes,
                **simulate_kwargs,
            )
        else:
            simulated_rf = ops.map(
                lambda inputs: simulate_rf(
                    scatterer_positions=inputs["positions"],
                    scatterer_magnitudes=inputs["magnitudes"],
                    **simulate_kwargs,
                ),
                {
                    "positions": scatterer_positions,
                    "magnitudes": scatterer_magnitudes,
                },
            )

        return {
            self.output_key: simulated_rf,
            "n_ch": 1,  # Simulate always returns RF data (so single channel)
        }


@ops_registry("tof_correction")
class TOFCorrection(Operation):
    """Time-of-flight correction operation for ultrasound data."""

    # Define operation-specific static parameters
    STATIC_PARAMS = ["f_number", "apply_lens_correction"]

    def __init__(self, **kwargs):
        super().__init__(
            input_data_type=DataTypes.RAW_DATA,
            output_data_type=DataTypes.ALIGNED_DATA,
            **kwargs,
        )

    def call(
        self,
        flatgrid,
        sound_speed,
        polar_angles,
        focus_distances,
        sampling_frequency,
        f_number,
        demodulation_frequency,
        t0_delays,
        tx_apodizations,
        initial_times,
        probe_geometry,
        t_peak,
        tx_waveform_indices,
        transmit_origins,
        apply_lens_correction=None,
        lens_thickness=None,
        lens_sound_speed=None,
        sos_map=None,
        sos_grid_x=None,
        sos_grid_z=None,
        **kwargs,
    ):
        """Perform time-of-flight correction on raw RF data.

        Args:
            raw_data (ops.Tensor): Raw RF data to correct
            flatgrid (ops.Tensor): Grid points at which to evaluate the time-of-flight
            sound_speed (float): Sound speed in the medium
            polar_angles (ops.Tensor): Polar angles for scan lines
            focus_distances (ops.Tensor): Focus distances for scan lines
            sampling_frequency (float): Sampling frequency
            f_number (float): F-number for apodization
            demodulation_frequency (float): Demodulation frequency
            t0_delays (ops.Tensor): T0 delays
            tx_apodizations (ops.Tensor): Transmit apodizations
            initial_times (ops.Tensor): Initial times
            probe_geometry (ops.Tensor): Probe element positions
            t_peak (float): Time to peak of the transmit pulse
            tx_waveform_indices (ops.Tensor): Index of the transmit waveform for each
                transmit. (All zero if there is only one waveform)
            transmit_origins (ops.Tensor): Transmit origins of shape (n_tx, 3)
            apply_lens_correction (bool): Whether to apply lens correction
            lens_thickness (float): Lens thickness
            lens_sound_speed (float): Sound speed in the lens
            sos_map (Tensor): Speed-of-sound map of shape ``(Nz, Nx)`` in m/s.
            sos_grid_x (Tensor): x-coordinates of ``sos_map`` rows.
            sos_grid_z (Tensor): z-coordinates of ``sos_map`` columns.

        Returns:
            dict: Dictionary containing tof_corrected_data
        """

        raw_data = kwargs[self.key]

        tof_kwargs = {
            "flatgrid": flatgrid,
            "t0_delays": t0_delays,
            "tx_apodizations": tx_apodizations,
            "sound_speed": sound_speed,
            "probe_geometry": probe_geometry,
            "initial_times": initial_times,
            "sampling_frequency": sampling_frequency,
            "demodulation_frequency": demodulation_frequency,
            "f_number": f_number,
            "polar_angles": polar_angles,
            "focus_distances": focus_distances,
            "t_peak": t_peak,
            "tx_waveform_indices": tx_waveform_indices,
            "transmit_origins": transmit_origins,
            "apply_lens_correction": apply_lens_correction,
            "lens_thickness": lens_thickness,
            "lens_sound_speed": lens_sound_speed,
            "sos_map": sos_map,
            "sos_grid_x": sos_grid_x,
            "sos_grid_z": sos_grid_z,
        }

        if not self.with_batch_dim:
            tof_corrected = tof_correction(raw_data, **tof_kwargs)
        else:
            tof_corrected = ops.map(
                lambda data: tof_correction(data, **tof_kwargs),
                raw_data,
            )

        return {self.output_key: tof_corrected}


@ops_registry("pfield_weighting")
class PfieldWeighting(Operation):
    """Weighting aligned data with the pressure field."""

    def __init__(self, **kwargs):
        super().__init__(
            input_data_type=DataTypes.ALIGNED_DATA,
            output_data_type=DataTypes.ALIGNED_DATA,
            **kwargs,
        )

    def call(self, flat_pfield=None, **kwargs):
        """Weight data with pressure field.

        Args:
            flat_pfield (ops.Tensor): Pressure field weight mask of shape (n_pix, n_tx)

        Returns:
            dict: Dictionary containing weighted data
        """
        data = kwargs[self.key]  # must start with ((batch_size,) n_tx, n_pix, ...)

        if flat_pfield is None:
            return {self.output_key: data}

        # Swap (n_pix, n_tx) to (n_tx, n_pix)
        flat_pfield = ops.swapaxes(flat_pfield, 0, 1)

        # Add batch dimension if needed
        if self.with_batch_dim:
            pfield_expanded = ops.expand_dims(flat_pfield, axis=0)
        else:
            pfield_expanded = flat_pfield

        append_n_dims = ops.ndim(data) - ops.ndim(pfield_expanded)
        pfield_expanded = extend_n_dims(pfield_expanded, axis=-1, n_dims=append_n_dims)

        # Perform element-wise multiplication with the pressure weight mask
        weighted_data = data * pfield_expanded

        return {self.output_key: weighted_data}


@ops_registry("scan_convert")
class ScanConvert(Operation):
    """Scan convert images to cartesian coordinates."""

    STATIC_PARAMS = ["fill_value"]
    ADD_OUTPUT_KEYS = [
        "resolution",
        "x_lim",
        "y_lim",
        "z_lim",
        "rho_range",
        "theta_range",
        "phi_range",
        "d_rho",
        "d_theta",
        "d_phi",
    ]

    def __init__(self, order=1, **kwargs):
        """Initialize the ScanConvert operation.

        Args:
            order (int, optional): Interpolation order. Defaults to 1. Currently only
                GPU support for order=1.
        """
        if order > 1:
            jittable = False
            log.warning(
                "GPU support for order > 1 is not available. " + "Disabling jit for ScanConvert."
            )
        else:
            jittable = True

        super().__init__(
            input_data_type=DataTypes.IMAGE,
            output_data_type=DataTypes.IMAGE_SC,
            jittable=jittable,
            **kwargs,
        )
        self.order = order

    def call(
        self,
        rho_range=None,
        theta_range=None,
        phi_range=None,
        resolution=None,
        coordinates=None,
        fill_value=None,
        **kwargs,
    ):
        """Scan convert images to cartesian coordinates.

        Args:
            rho_range (Tuple): Range of the rho axis in the polar coordinate system.
                Defined in meters.
            theta_range (Tuple): Range of the theta axis in the polar coordinate system.
                Defined in radians.
            phi_range (Tuple): Range of the phi axis in the polar coordinate system.
                Defined in radians.
            resolution (float): Resolution of the output image in meters per pixel.
                if None, the resolution is computed based on the input data.
            coordinates (Tensor): Coordinates for scan convertion. If None, will be computed
                based on rho_range, theta_range, phi_range and resolution. If provided, this
                operation can be jitted.
            fill_value (float): Value to fill the image with outside the defined region.

        """
        if fill_value is None:
            fill_value = np.nan

        data = kwargs[self.key]

        if self._jit_compile and self.jittable:
            assert coordinates is not None, (
                "coordinates must be provided to jit scan conversion."
                "You can set ScanConvert(jit_compile=False) to disable jitting."
            )

        data_out, parameters = scan_convert(
            data,
            rho_range,
            theta_range,
            phi_range,
            resolution,
            coordinates,
            fill_value,
            self.order,
            with_batch_dim=self.with_batch_dim,
        )

        return {self.output_key: data_out, **parameters}


@ops_registry("demodulate")
class Demodulate(Operation):
    """Demodulates the input data to baseband. After this operation, the carrier frequency
    is removed (0 Hz) and the data is in IQ format stored in two real valued channels."""

    ADD_OUTPUT_KEYS = ["center_frequency", "n_ch"]

    def __init__(self, axis=-3, **kwargs):
        super().__init__(
            input_data_type=DataTypes.RAW_DATA,
            output_data_type=DataTypes.RAW_DATA,
            jittable=True,
            **kwargs,
        )
        self.axis = axis

    def call(self, demodulation_frequency=None, sampling_frequency=None, **kwargs):
        data = kwargs[self.key]

        # Split the complex signal into two channels
        iq_data_two_channel = demodulate(
            data=data,
            demodulation_frequency=demodulation_frequency,
            sampling_frequency=sampling_frequency,
            axis=self.axis,
        )

        return {
            self.output_key: iq_data_two_channel,
            "center_frequency": 0.0,
            "n_ch": 2,
        }


@ops_registry("fir_filter")
class FirFilter(Operation):
    """Apply a FIR filter to the input signal using convolution.

    Looks for the filter taps in the input dictionary using the specified ``filter_key``.
    """

    def __init__(
        self,
        axis: int,
        complex_channels: bool = False,
        filter_key: str = "fir_filter_taps",
        **kwargs,
    ):
        """
        Args:
            axis (int): Axis along which to apply the filter. Cannot be the batch dimension and
                not the complex channel axis when ``complex_channels=True``.
            complex_channels (bool): Whether the last dimension of the input signal represents
                complex channels (real and imaginary parts). When True, it will convert the signal
                to ``complex`` dtype before filtering and convert it back to two channels
                after filtering.
            filter_key (str): Key in the input dictionary where the FIR filter taps are stored.
                Default is "fir_filter_taps".
        """
        super().__init__(**kwargs)
        self._check_axis(axis)

        self.axis = axis
        self.complex_channels = complex_channels
        self.filter_key = filter_key

    def _check_axis(self, axis, ndim=None):
        """Check if axis is not the batch dimension."""
        if self.with_batch_dim and (axis == 0 or (ndim is not None and axis == -ndim)):
            raise ValueError("Cannot apply FIR filter along batch dimension.")

    @property
    def valid_keys(self):
        """Get the valid keys for the `call` method."""
        return self._valid_keys.union({self.filter_key})

    def call(self, **kwargs):
        signal = kwargs[self.key]
        fir_filter_taps = kwargs[self.filter_key]

        ndim = ops.ndim(signal)
        self._check_axis(self.axis, ndim)
        axis = canonicalize_axis(self.axis, ndim)

        if self.complex_channels:
            assert axis < ndim - 1, (
                "When using complex_channels=True, the complex channels are removed to convert"
                " to complex numbers before filtering, so axis cannot be the last axis."
            )
            signal = channels_to_complex(signal)

        def _convolve(signal):
            """Apply the filter to the signal using correlation."""
            return correlate(signal, ops.flip(fir_filter_taps, axis=0), mode="same")

        filtered_signal = apply_along_axis(_convolve, axis, signal)

        if self.complex_channels:
            filtered_signal = complex_to_channels(filtered_signal)

        return {self.output_key: filtered_signal}


@ops_registry("low_pass_filter")
class LowPassFilterIQ(FirFilter):
    """Apply a low-pass FIR filter to the demodulated IQ (n_ch=2) input signal using convolution.

    It is recommended to use :class:`FirFilter` with pre-computed filter taps for jittable
    operations. The :class:`LowPassFilterIQ` operation itself is not jittable and is provided
    for convenience only.

    Uses :func:`get_low_pass_iq_filter` to compute the filter taps.
    """

    def __init__(
        self, axis: int = -3, num_taps: int = 127, filter_key: str = "low_pass_filter", **kwargs
    ):
        """Initialize the LowPassFilterIQ operation.

        Args:
            axis (int): Axis along which to apply the filter. Cannot be the batch dimension and
                cannot be the complex channel axis (the last axis). Default is -3, which is the
                ``n_ax`` axis for standard ultrasound data layout.
            num_taps (int): Number of taps in the FIR filter. Default is 127.
                Odd will result in a type I filter, even in a type II filter.
        """
        if "jittable" in kwargs:
            raise ValueError("LowPassFilterIQ is not jittable, so jittable must be set to False.")
        if "complex_channels" in kwargs and not kwargs["complex_channels"]:
            raise ValueError(
                "LowPassFilterIQ operates on IQ data, so complex_channels must be True."
            )
        kwargs.pop("jittable", None)
        kwargs.pop("complex_channels", None)
        super().__init__(
            axis=axis,
            complex_channels=True,
            filter_key=filter_key,
            jittable=False,
            **kwargs,
        )
        self.num_taps = num_taps

    def call(self, bandwidth, sampling_frequency, center_frequency, **kwargs):
        lpf = get_low_pass_iq_filter(
            self.num_taps,
            ops.convert_to_numpy(sampling_frequency).item(),
            ops.convert_to_numpy(center_frequency).item(),
            ops.convert_to_numpy(bandwidth).item(),
        )
        kwargs[self.filter_key] = lpf
        return super().call(**kwargs)


@ops_registry("band_pass_filter")
class BandPassFilter(FirFilter):
    """Apply a band-pass FIR filter to the real input signal using convolution.

    The bandwidth parameter in the call method defines the passband centered around
    ``demodulation_frequency``, with edges at ``demodulation_frequency - bandwidth/2``
    and ``demodulation_frequency + bandwidth/2``. So, make sure this is used before demodulation
    to baseband.

    This operation is provided for convenience and will recompute the filter weights every
    time it is called. Alternatively, you can use :class:`FirFilter` with pre-computed
    filter taps.
    """

    def __init__(
        self, axis: int = -3, num_taps: int = 127, filter_key: str = "band_pass_filter", **kwargs
    ):
        """Initialize the BandPassFilter operation.

        Args:
            axis (int): Axis along which to apply the filter. Cannot be the batch dimension.
                Default is -3, which is the ``n_ax`` axis for standard ultrasound data layout.
            num_taps (int): Number of taps in the FIR filter. Default is 127.
                Odd will result in a type I filter, even in a type II filter.
        """
        if "complex_channels" in kwargs and kwargs["complex_channels"]:
            raise ValueError(
                "BandPassFilter operates on a real signal, so complex_channels must be False."
            )
        kwargs.pop("complex_channels", None)
        super().__init__(
            axis=axis,
            complex_channels=False,
            filter_key=filter_key,
            **kwargs,
        )
        self.num_taps = num_taps

    def call(self, sampling_frequency, demodulation_frequency, bandwidth, **kwargs):
        """Apply band-pass filter with specified bandwidth.

        Args:
            sampling_frequency (float): Sampling frequency in Hz.
            demodulation_frequency (float): Center frequency in Hz.
            bandwidth (float): Bandwidth in Hz. The filter will pass frequencies from
                ``demodulation_frequency - bandwidth/2`` to
                ``demodulation_frequency + bandwidth/2``.

        Returns:
            dict: Dictionary containing filtered signal.
        """
        f1 = demodulation_frequency - bandwidth / 2
        f2 = demodulation_frequency + bandwidth / 2

        bpf = get_band_pass_filter(
            self.num_taps, sampling_frequency, f1, f2, validate=not self._jit_compile
        )
        kwargs[self.filter_key] = bpf
        return super().call(**kwargs)


@ops_registry("channels_to_complex")
class ChannelsToComplex(Operation):
    def call(self, **kwargs):
        data = kwargs[self.key]
        output = channels_to_complex(data)
        return {self.output_key: output}


@ops_registry("complex_to_channels")
class ComplexToChannels(Operation):
    def __init__(self, axis=-1, **kwargs):
        super().__init__(**kwargs)
        self.axis = axis

    def call(self, **kwargs):
        data = kwargs[self.key]
        output = complex_to_channels(data, axis=self.axis)
        return {self.output_key: output}


@ops_registry("lee_filter")
class LeeFilter(Filter):
    """
    The Lee filter is a speckle reduction filter commonly used in synthetic aperture radar (SAR)
    and ultrasound image processing. It smooths the image while preserving edges and details.
    This implementation uses Gaussian filter for local statistics and treats channels independently.

    Lee, J.S. (1980). Digital image enhancement and noise filtering by use of local statistics.
    IEEE Transactions on Pattern Analysis and Machine Intelligence, (2), 165-168.
    """

    def __init__(
        self,
        sigma: float,
        mode: str = "symmetric",
        cval: float | None = None,
        truncate: float = 4.0,
        axes: Tuple[int] = (-3, -2),
        **kwargs,
    ):
        """
        Args:
            sigma (float or tuple): Standard deviation for Gaussian kernel. The standard deviations
                of the Gaussian filter are given for each axis as a sequence, or as a single number,
                in which case it is equal for all axes.
            mode (str, optional): Padding mode for the input image. Default is 'symmetric'.
                See [keras docs](https://www.tensorflow.org/api_docs/python/tf/keras/ops/pad) for
                all options and [tensorflow docs](https://www.tensorflow.org/api_docs/python/tf/pad)
                for some examples. Note that the naming differs from scipy.ndimage.gaussian_filter!
            cval (float, optional): Value to fill past edges of input if mode is 'constant'.
                Default is None.
            truncate (float, optional): Truncate the filter at this many standard deviations.
                Default is 4.0.
            axes (Tuple[int], optional): If None, input is filtered along all axes. Otherwise, input
                is filtered along the specified axes. When axes is specified, any tuples used for
                sigma, order, mode and/or radius must match the length of axes. The ith entry in
                any of these tuples corresponds to the ith entry in axes. Default is (-3, -2),
                which corresponds to the height and width dimensions of a
                (..., height, width, channels) tensor.
        """
        super().__init__(**kwargs)
        self.sigma = sigma
        self.mode = mode
        self.cval = cval
        self.truncate = truncate
        self.axes = axes

    def call(self, **kwargs):
        """Apply the Lee filter to the input data.

        Args:
            data (ops.Tensor): Input image data of shape (height, width, channels) with
                optional batch dimension if ``self.with_batch_dim``.
        """
        data = kwargs.pop(self.key)
        axes = self._resolve_filter_axes(data, self.axes)

        # Apply Gaussian blur to get local mean
        img_mean = gaussian_filter(
            data, self.sigma, mode=self.mode, cval=self.cval, truncate=self.truncate, axes=axes
        )

        # Apply Gaussian blur to squared data to get local squared mean
        img_sqr_mean = gaussian_filter(
            data**2, self.sigma, mode=self.mode, cval=self.cval, truncate=self.truncate, axes=axes
        )

        # Calculate local variance
        img_variance = img_sqr_mean - img_mean**2

        # Calculate global variance (per channel)
        overall_variance = ops.var(data, axis=axes, keepdims=True)

        # Calculate adaptive weights
        eps = keras.config.epsilon()
        img_weights = img_variance / (img_variance + overall_variance + eps)

        # Apply Lee filter formula
        img_output = img_mean + img_weights * (data - img_mean)

        return {self.output_key: img_output}


@ops_registry("companding")
class Companding(Operation):
    """Companding according to the A- or μ-law algorithm.

    Invertible compressing operation. Used to compress
    dynamic range of input data (and subsequently expand).

    μ-law companding:
    https://en.wikipedia.org/wiki/%CE%9C-law_algorithm
    A-law companding:
    https://en.wikipedia.org/wiki/A-law_algorithm

    Args:
        expand (bool, optional): If set to False (default),
            data is compressed, else expanded.
        comp_type (str): either `a` or `mu`.
        mu (float, optional): compression parameter. Defaults to 255.
        A (float, optional): compression parameter. Defaults to 87.6.
    """

    def __init__(self, expand=False, comp_type="mu", **kwargs):
        super().__init__(**kwargs)
        self.expand = expand
        self.comp_type = comp_type.lower()
        if self.comp_type not in ["mu", "a"]:
            raise ValueError("comp_type must be 'mu' or 'a'.")

        if self.comp_type == "mu":
            self._compand_func = self._mu_law_expand if self.expand else self._mu_law_compress
        else:
            self._compand_func = self._a_law_expand if self.expand else self._a_law_compress

    @staticmethod
    def _mu_law_compress(x, mu=255, **kwargs):
        x = ops.clip(x, -1, 1)
        return ops.sign(x) * ops.log(1.0 + mu * ops.abs(x)) / ops.log(1.0 + mu)

    @staticmethod
    def _mu_law_expand(y, mu=255, **kwargs):
        y = ops.clip(y, -1, 1)
        return ops.sign(y) * ((1.0 + mu) ** ops.abs(y) - 1.0) / mu

    @staticmethod
    def _a_law_compress(x, A=87.6, **kwargs):
        x = ops.clip(x, -1, 1)
        x_sign = ops.sign(x)
        x_abs = ops.abs(x)
        A_log = ops.log(A)
        val1 = x_sign * A * x_abs / (1.0 + A_log)
        val2 = x_sign * (1.0 + ops.log(A * x_abs)) / (1.0 + A_log)
        y = ops.where((x_abs >= 0) & (x_abs < (1.0 / A)), val1, val2)
        return y

    @staticmethod
    def _a_law_expand(y, A=87.6, **kwargs):
        y = ops.clip(y, -1, 1)
        y_sign = ops.sign(y)
        y_abs = ops.abs(y)
        A_log = ops.log(A)
        val1 = y_sign * y_abs * (1.0 + A_log) / A
        val2 = y_sign * ops.exp(y_abs * (1.0 + A_log) - 1.0) / A
        x = ops.where((y_abs >= 0) & (y_abs < (1.0 / (1.0 + A_log))), val1, val2)
        return x

    def call(self, mu=255, A=87.6, **kwargs):
        data = kwargs[self.key]

        mu = ops.cast(mu, data.dtype)
        A = ops.cast(A, data.dtype)

        data_out = self._compand_func(data, mu=mu, A=A)
        return {self.output_key: data_out}


@ops_registry("downsample")
class Downsample(Operation):
    """Downsample data along a specific axis."""

    ADD_OUTPUT_KEYS = ["sampling_frequency", "n_ax"]

    def __init__(self, factor: int = 1, phase: int = 0, axis: int = -3, **kwargs):
        super().__init__(
            **kwargs,
        )
        if factor < 1:
            raise ValueError("Downsample factor must be >= 1.")
        if phase < 0 or phase >= factor:
            raise ValueError("phase must satisfy 0 <= phase < factor.")
        self.factor = factor
        self.phase = phase
        self.axis = axis

    def call(self, sampling_frequency=None, n_ax=None, **kwargs):
        data = kwargs[self.key]
        length = ops.shape(data)[self.axis]
        sample_idx = ops.arange(self.phase, length, self.factor)
        data_downsampled = ops.take(data, sample_idx, axis=self.axis)

        output = {self.output_key: data_downsampled}
        # downsampling also affects the sampling frequency
        if sampling_frequency is not None:
            sampling_frequency = sampling_frequency / self.factor
            output["sampling_frequency"] = sampling_frequency
        if n_ax is not None:
            n_ax = n_ax // self.factor
            output["n_ax"] = n_ax
        return output


@ops_registry("anisotropic_diffusion")
class AnisotropicDiffusion(Operation):
    """Speckle Reducing Anisotropic Diffusion (SRAD) filter.

    Reference:
    - https://www.researchgate.net/publication/5602035_Speckle_reducing_anisotropic_diffusion
    - https://nl.mathworks.com/matlabcentral/fileexchange/54044-image-despeckle-filtering-toolbox
    """

    def call(self, niter=100, lmbda=0.1, rect=None, eps=1e-6, **kwargs):
        """Anisotropic diffusion filter.

        Assumes input data is non-negative.

        Args:
            niter: Number of iterations.
            lmbda: Lambda parameter.
            rect: Rectangle [x1, y1, x2, y2] for homogeneous noise (optional).
            eps: Small epsilon for stability.
        Returns:
            Filtered image (2D tensor or batch of images).
        """
        data = kwargs[self.key]

        if not self.with_batch_dim:
            data = ops.expand_dims(data, axis=0)

        batch_size = ops.shape(data)[0]

        results = []
        for i in range(batch_size):
            image = data[i]
            image_out = self._anisotropic_diffusion_single(image, niter, lmbda, rect, eps)
            results.append(image_out)

        result = ops.stack(results, axis=0)

        if not self.with_batch_dim:
            result = ops.squeeze(result, axis=0)

        return {self.output_key: result}

    def _anisotropic_diffusion_single(self, image, niter, lmbda, rect, eps):
        """Apply anisotropic diffusion to a single image (2D)."""
        image = ops.exp(image)
        M, N = image.shape

        for _ in range(niter):
            iN = ops.concatenate([image[1:], ops.zeros((1, N), dtype=image.dtype)], axis=0)
            iS = ops.concatenate([ops.zeros((1, N), dtype=image.dtype), image[:-1]], axis=0)
            jW = ops.concatenate([image[:, 1:], ops.zeros((M, 1), dtype=image.dtype)], axis=1)
            jE = ops.concatenate([ops.zeros((M, 1), dtype=image.dtype), image[:, :-1]], axis=1)

            if rect is not None:
                x1, y1, x2, y2 = rect
                imageuniform = image[x1:x2, y1:y2]
                q0_squared = (ops.std(imageuniform) / (ops.mean(imageuniform) + eps)) ** 2

            dN = iN - image
            dS = iS - image
            dW = jW - image
            dE = jE - image

            G2 = (dN**2 + dS**2 + dW**2 + dE**2) / (image**2 + eps)
            L = (dN + dS + dW + dE) / (image + eps)
            num = (0.5 * G2) - ((1 / 16) * (L**2))
            den = (1 + ((1 / 4) * L)) ** 2
            q_squared = num / (den + eps)

            if rect is not None:
                den = (q_squared - q0_squared) / (q0_squared * (1 + q0_squared) + eps)
            c = 1.0 / (1 + den)
            cS = ops.concatenate([ops.zeros((1, N), dtype=image.dtype), c[:-1]], axis=0)
            cE = ops.concatenate([ops.zeros((M, 1), dtype=image.dtype), c[:, :-1]], axis=1)

            D = (cS * dS) + (c * dN) + (cE * dE) + (c * dW)
            image = image + (lmbda / 4) * D

        result = ops.log(image)
        return result


@ops_registry("envelope_detect")
class EnvelopeDetect(Operation):
    """Envelope detection of RF signals."""

    def __init__(
        self,
        axis=-3,
        **kwargs,
    ):
        super().__init__(
            input_data_type=DataTypes.BEAMFORMED_DATA,
            output_data_type=DataTypes.ENVELOPE_DATA,
            **kwargs,
        )
        self.axis = axis

    def call(self, **kwargs):
        """
        Args:
            - data (Tensor): The beamformed data of shape (..., grid_size_z, grid_size_x, n_ch).
        Returns:
            - envelope_data (Tensor): The envelope detected data
                of shape (..., grid_size_z, grid_size_x).
        """
        data = kwargs[self.key]

        data = envelope_detect(data, axis=self.axis)

        return {self.output_key: data}


@ops_registry("upmix")
class UpMix(Operation):
    """Upmix IQ data to RF data."""

    def __init__(
        self,
        upsampling_rate=1,
        **kwargs,
    ):
        super().__init__(
            **kwargs,
        )
        self.upsampling_rate = upsampling_rate

    def call(self, sampling_frequency=None, demodulation_frequency=None, **kwargs):
        data = kwargs[self.key]

        if data.shape[-1] == 1:
            log.warning("Upmixing is not applicable to RF data.")
            return {self.output_key: data}
        elif data.shape[-1] == 2:
            data = channels_to_complex(data)

        data = upmix(data, sampling_frequency, demodulation_frequency, self.upsampling_rate)
        data = ops.expand_dims(data, axis=-1)
        return {self.output_key: data}


@ops_registry("log_compress")
class LogCompress(Operation):
    """Logarithmic compression of data."""

    def __init__(self, clip: bool = True, **kwargs):
        """Initialize the LogCompress operation.

        Args:
            clip (bool): Whether to clip the output to a dynamic range. Defaults to True.
        """
        super().__init__(
            input_data_type=DataTypes.ENVELOPE_DATA,
            output_data_type=DataTypes.IMAGE,
            **kwargs,
        )
        self.clip = clip

    def call(self, dynamic_range=None, **kwargs):
        """Apply logarithmic compression to data.

        Args:
            dynamic_range (tuple, optional): Dynamic range in dB. Defaults to (-60, 0).

        Returns:
            dict: Dictionary containing log-compressed data
        """
        data = kwargs[self.key]

        if dynamic_range is None:
            dynamic_range = ops.array(DEFAULT_DYNAMIC_RANGE)
        dynamic_range = ops.cast(dynamic_range, data.dtype)

        compressed_data = log_compress(data)
        if self.clip:
            compressed_data = ops.clip(compressed_data, dynamic_range[0], dynamic_range[1])

        return {self.output_key: compressed_data}


@ops_registry("reshape_grid")
class ReshapeGrid(Operation):
    """Reshape flat grid data to grid shape."""

    def __init__(self, axis=0, **kwargs):
        super().__init__(**kwargs)
        self.axis = axis

    def call(self, grid, **kwargs):
        """
        Args:
            - data (Tensor): The flat grid data of shape (..., n_pix, ...).
        Returns:
            - reshaped_data (Tensor): The reshaped data of shape (..., grid.shape, ...).
        """
        data = kwargs[self.key]
        reshaped_data = reshape_axis(data, grid.shape[:-1], self.axis + int(self.with_batch_dim))
        return {self.output_key: reshaped_data}


@ops_registry("apply_window")
class ApplyWindow(Operation):
    """Apply a window function to the input data along a specific axis.

    This operation can be used to zero out the end and/or beginning of the signal and apply a window
    of some size to transition from the zeroed region to the unmodified region.

    The axis is divided into five regions:
    [start (zero)] - [size (window)] - [middle (unmodified)] - [size (window)] - [end (zero)]
    """

    STATIC_PARAMS = ["axis", "size", "window_type", "start", "end"]

    def __init__(self, axis=-3, size=32, start=16, end=0, window_type="hanning", **kwargs):
        """
        Args:
            axis (int): Axis along which to apply the window.
            size (int): Size of the window to apply at the start and end regions.
            start (int): Number of elements to zero at the end.
            end (int): Number of elements to zero at the end.
            window_type (str): Type of window to apply. Supported types are "hanning" and "linear".
        """
        super().__init__(**kwargs)
        self.axis = axis
        self.size = int(size)
        self.start = int(start)
        self.end = int(end)
        self._check_inputs()
        self.window_type = window_type
        self.window = self._get_window(self.window_type, size, "float32")

    def _check_inputs(self):
        if self.start < 0:
            raise ValueError("start must be >= 0.")
        if self.end < 0:
            raise ValueError("end must be >= 0.")
        if self.size < 0:
            raise ValueError("size must be >= 0.")

    @staticmethod
    def _get_window(window_type, size, dtype):
        if window_type == "hanning":
            window = ops.hanning(size * 2)
        elif window_type == "linear":
            window = ops.concatenate(
                [ops.linspace(0.0, 1.0, size), ops.linspace(1.0, 0.0, size)], axis=0
            )
        else:
            raise ValueError(f"Unsupported window type: {window_type}")
        return ops.cast(window, dtype)

    def call(self, **kwargs):
        data = kwargs[self.key]
        dtype = ops.dtype(data)
        axis = canonicalize_axis(self.axis, ops.ndim(data))

        length = ops.shape(data)[axis]

        if self.start + self.size * 2 + self.end > length:
            raise ValueError("start, size, and end are larger than the axis length.")

        window = ops.cast(self.window, dtype)

        ones = ops.ones((length,), dtype=dtype)
        mask = ops.concatenate(
            [
                ops.zeros((self.start,), dtype=dtype),
                window[: self.size],
                ones[self.size + self.start : -(self.end + self.size)],
                window[self.size :],
                ops.zeros((self.end,), dtype=dtype),
            ],
            axis=0,
        )

        shape = [1] * ops.ndim(data)
        shape[axis] = length
        mask = ops.reshape(mask, shape)

        return {self.output_key: data * mask}


@ops_registry("common_midpoint_phase_error")
class CommonMidpointPhaseError(Operation):
    """Calculates the Common Midpoint Phase Error (CMPE)

    Computes CMPE between translated transmit and receive apertures with a common midpoint.

    .. important::
        Only works for multistatic datasets, e.g. synthetic aperture data.

    .. note::
        This was directly adapted from the Differentiable Beamforming for Ultrasound Autofocusing (DBUA)
        paper, see `original paper and code <https://waltersimson.com/dbua/>`_.

    """  # noqa: E501

    def _init_(
        self,
        reshape_grid=True,
        **kwargs,
    ):
        super()._init_(
            input_data_type=None,
            # DataTypes.IMAGE, because we have an image of the phase map
            output_data_type=DataTypes.IMAGE,
            **kwargs,
        )
        self.reshape_grid = reshape_grid

    def create_subapertures(self, data, halfsa, dx):
        """Create subapertures from the data.

        Args:
            data (ops.Tensor): The data to create subapertures from.
            halfsa (int): Half of the subaperture.
            dx (float): The spacing between the subapertures.

        Returns:
            transmit_subap (ops.Tensor): The transmit subapertures.
            receive_subap (ops.Tensor): The receive subapertures.
        """
        n_tx, n_pix, n_rx, n_ch = data.shape
        receive_subaps = ops.zeros((n_rx, n_tx))
        for diag in range(-halfsa, halfsa + 1):
            receive_subaps = receive_subaps + ops.diag(ops.ones((n_rx - abs(diag),)), diag)
        receive_subaps = receive_subaps[halfsa : receive_subaps.shape[0] - halfsa : dx]
        transmit_subaps = ops.flip(receive_subaps, axis=0)
        return transmit_subaps, receive_subaps

    def process_phase_map(self, data, **kwargs):
        """Create the common midpoint subaperture phase error map.

        Args:
            data (ops.Tensor): The data to create the phase error map from.

        Returns:
            phase_error_map (ops.Tensor): The phase error map.
        """

        transmit_subaps, receive_subaps = self.create_subapertures(data, 8, 1)
        complex_data = ops.view_as_complex(data)  # [n_tx, n_pix, n_rx, n_ch] -> [n_rtx, n_pix, r_x]
        complex_data = ops.transpose(complex_data, (2, 0, 1))  # [n_rx, n_tx, n_pix]
        rx_zero_count = ops.matmul(receive_subaps, ops.cast(complex_data == 0, "int32"))

        # Mask out subapertures with point outside fov in receive
        rx_valid = rx_zero_count <= 1
        complex_data_rx = ops.matmul(receive_subaps, complex_data)
        complex_data_rx = ops.where(rx_valid, complex_data_rx, 0)
        complex_data_rx = ops.transpose(complex_data_rx, (1, 0, 2))  # [n_tx, n_subap_rx, n_pix]
        tx_zero_count = ops.matmul(transmit_subaps, ops.cast(complex_data_rx == 0, "int32"))

        # Mask out subapertures with point outside fov in transmit
        tx_valid = tx_zero_count <= 1

        data = ops.matmul(transmit_subaps, complex_data_rx)
        data = ops.where(tx_valid, data, 0)
        data = ops.transpose(data, (1, 0, 2))  # [n_subap_tx, n_subap, n_pix]

        # take diagonals
        a = data[:-1, :-1]
        b = data[1:, 1:]
        valid = (a != 0) & (b != 0)

        # compute phase difference between cmp neighbours
        # This only works if the array is regularly spaced
        xy = a * ops.conj(b)
        xy = ops.where(valid, xy, 0)
        dphi = ops.angle(xy)
        dphi = ops.abs(dphi)

        dphi = ops.sum(dphi, (0, 1)) / ops.cast(ops.sum(valid, (0, 1)), dphi.dtype)
        return dphi

    def call(
        self,
        **kwargs,
    ):
        data = kwargs[self.key]
        if not self.with_batch_dim:
            pemap = self.process_phase_map(data)
        else:
            pemap = ops.map(
                lambda d: self.process_phase_map(d),
                data,
            )
        return {self.output_key: pemap}



"""Input / output functions for reading and writing files.

Use to quickly read and write files or interact with file system.
"""

import functools
import os
import time
from io import BytesIO
from pathlib import Path
from typing import Generator

import imageio
import numpy as np
from PIL import Image, ImageSequence

from zea import log

_SUPPORTED_VID_TYPES = [".mp4", ".gif"]
_SUPPORTED_IMG_TYPES = [".jpg", ".png", ".JPEG", ".PNG", ".jpeg"]
_SUPPORTED_ZEA_TYPES = [".hdf5", ".h5"]


def load_video(filename, mode="L"):
    """Load a video file and return a numpy array of frames.

    Supported file types: mp4, gif.

    Args:
        filename (str): The path to the video file.
        mode (str, optional): Color mode: "L" (grayscale) or "RGB".
            Defaults to "L".

    Returns:
        numpy.ndarray: Array of frames (num_frames, H, W) or (num_frames, H, W, C)

    Raises:
        ValueError: If the file extension or mode is not supported.
    """
    filename = Path(filename)
    if not filename.exists():
        raise FileNotFoundError(f"File {filename} does not exist")
    ext = filename.suffix.lower()

    if ext not in _SUPPORTED_VID_TYPES:
        raise ValueError(f"Unsupported file extension: {ext}")

    if mode not in {"L", "RGB"}:
        raise ValueError(f"Unsupported mode: {mode}")

    frames = []

    if ext == ".gif":
        with Image.open(filename) as im:
            for frame in ImageSequence.Iterator(im):
                frames.append(_convert_image_mode(frame, mode=mode))
    elif ext == ".mp4":
        # Use imageio with FFMPEG format for MP4 files
        try:
            reader = imageio.get_reader(filename, format="FFMPEG")
        except (ImportError, ValueError) as exc:
            raise ImportError(
                "FFMPEG plugin is required to load MP4 files. "
                "Please install it with 'pip install imageio-ffmpeg'."
            ) from exc

        try:
            for frame in reader:
                img = Image.fromarray(frame)
                frames.append(_convert_image_mode(img, mode=mode))
        finally:
            reader.close()
    else:
        raise ValueError(f"Unsupported file extension: {ext}")

    return np.stack(frames, axis=0)


def load_image(filename, mode="L"):
    """Load an image file and return a numpy array.

    Supported file types: jpg, png.

    Args:
        filename (str): The path to the image file.
        mode (str, optional): Color mode: "L" (grayscale) or "RGB".
            Defaults to "L".

    Returns:
        numpy.ndarray: A numpy array of the image.

    Raises:
        ValueError: If the file extension or mode is not supported.
    """
    filename = Path(filename)
    if not filename.exists():
        raise FileNotFoundError(f"File {filename} does not exist")
    extension = filename.suffix.lower()
    allowed_exts = {ext.lower() for ext in _SUPPORTED_IMG_TYPES}
    if extension not in allowed_exts:
        raise ValueError(f"File extension {extension} not supported")

    if mode not in {"L", "RGB"}:
        raise ValueError(f"Unsupported mode: {mode}")

    with Image.open(filename) as img:
        return _convert_image_mode(img, mode=mode)


def save_video(images, filename, fps=20, **kwargs):
    """Saves a sequence of images to a video file.

    Supported file types: mp4, gif.

    Args:
        images (list or np.ndarray): List or array of images. Must have shape
            (n_frames, height, width, channels) or (n_frames, height, width).
            If channel axis is not present, or is 1, grayscale image is assumed,
            which is then converted to RGB. Images should be uint8.
        filename (str or Path): Filename to which data should be written.
        fps (int): Frames per second of rendered format.
        **kwargs: Additional keyword arguments passed to the specific save function.
            For GIF and mp4 files, this includes `shared_color_palette` (bool).

    Raises:
        ValueError: If the file extension is not supported.

    """
    filename = Path(filename)
    ext = filename.suffix.lower()

    if ext == ".mp4":
        return save_to_mp4(images, filename, fps=fps, **kwargs)
    elif ext == ".gif":
        return save_to_gif(images, filename, fps=fps, **kwargs)
    else:
        raise ValueError(f"Unsupported file extension: {ext}")


def save_to_gif(images, filename, fps=20, shared_color_palette=True):
    """Saves a sequence of images to a GIF file.

    .. note::
        It's recommended to use :func:`save_video` for a more general interface
        that supports multiple formats.

    Args:
        images (list or np.ndarray): List or array of images. Must have shape
            (n_frames, height, width, channels) or (n_frames, height, width).
            If channel axis is not present, or is 1, grayscale image is assumed,
            which is then converted to RGB. Images should be uint8.
        filename (str or Path): Filename to which data should be written.
        fps (int): Frames per second of rendered format.
        shared_color_palette (bool, optional): If True, creates a global
            color palette across all frames, ensuring consistent colors
            throughout the GIF. Defaults to True, which is default behavior
            of PIL.Image.save. Note: True increases speed and shrinks file
            size for longer sequences.

    """
    images = preprocess_for_saving(images)

    if fps > 50:
        log.warning(f"Cannot set fps ({fps}) > 50. Setting it automatically to 50.")
        fps = 50

    duration = int(round(1000 / fps))  # milliseconds per frame

    pillow_imgs = [Image.fromarray(img) for img in images]

    if shared_color_palette:
        # Apply the same palette to all frames without dithering for consistent color mapping
        global_palette = compute_global_palette_by_histogram(
            pillow_imgs, bits_per_channel=5, palette_size=256
        )

        # Apply the same palette to all frames without dithering
        pillow_imgs = [
            img.convert("RGB").quantize(
                palette=global_palette,
                dither=Image.NONE,
            )
            for img in pillow_imgs
        ]

    pillow_img, *pillow_imgs = pillow_imgs

    pillow_img.save(
        fp=filename,
        format="GIF",
        append_images=pillow_imgs,
        save_all=True,
        loop=0,
        duration=duration,
        interlace=False,
        optimize=False,
    )
    log.success(f"Successfully saved GIF to -> {log.yellow(filename)}")


def save_to_mp4(images, filename, fps=20, shared_color_palette=False):
    """Saves a sequence of images to an MP4 file.

    .. note::
        It's recommended to use :func:`save_video` for a more general interface
        that supports multiple formats.

    Args:
        images (list or np.ndarray): List or array of images. Must have shape
            (n_frames, height, width, channels) or (n_frames, height, width).
            If channel axis is not present, or is 1, grayscale image is assumed,
            which is then converted to RGB. Images should be uint8.
        filename (str or Path): Filename to which data should be written.
        fps (int): Frames per second of rendered format.
        shared_color_palette (bool, optional): If True, creates a global
            color palette across all frames, ensuring consistent colors
            throughout the MP4. Note: True can cause slow saving for longer
            sequences.

    Raises:
        ImportError: If imageio-ffmpeg is not installed.

    Returns:
        str: Success message.

    """
    images = preprocess_for_saving(images)

    filename = str(filename)

    parent_dir = Path(filename).parent
    parent_dir.mkdir(parents=True, exist_ok=True)

    # Use imageio with FFMPEG format for MP4 files
    try:
        writer = imageio.get_writer(
            filename, fps=fps, format="FFMPEG", codec="libx264", pixelformat="yuv420p"
        )
    except (ImportError, ValueError) as exc:
        raise ImportError(
            "FFMPEG plugin is required to save MP4 files. "
            "Please install it with 'pip install imageio-ffmpeg'."
        ) from exc

    try:
        if shared_color_palette:
            pillow_imgs = [Image.fromarray(img) for img in images]
            global_palette = compute_global_palette_by_histogram(
                pillow_imgs, bits_per_channel=5, palette_size=256
            )
            for img in pillow_imgs:
                paletted_img = img.convert("RGB").quantize(
                    palette=global_palette,
                    dither=Image.NONE,
                )
                writer.append_data(np.array(paletted_img.convert("RGB")))
        else:
            # Write from numpy arrays directly
            for image in images:
                writer.append_data(image)
    finally:
        writer.close()

    return log.success(f"Successfully saved MP4 to -> {filename}")


def search_file_tree(directory, filetypes=None, verbose=True, relative=False) -> Generator:
    """Traverse a directory tree and yield file paths matching specified file types.

    Args:
        directory (str or Path): The root directory to start the search.
        filetypes (list of str, optional): List of file extensions to match.
            If None, file types supported by `zea` are matched. Defaults to None.
        verbose (bool, optional): If True, logs the search process. Defaults to True.
        relative (bool, optional): If True, yields file paths relative to the
            root directory. Defaults to False.

    Yields:
        Path: Paths of files matching the specified file types.
    """
    # Traverse file tree to index all files from filetypes
    if verbose:
        log.info(f"Searching {log.yellow(directory)} for {filetypes} files...")

    for dirpath, _, filenames in os.walk(directory):
        for file in filenames:
            # Append to file_paths if it is a filetype file
            if Path(file).suffix in filetypes:
                file_path = Path(dirpath) / file
                if relative:
                    file_path = file_path.relative_to(directory)
                yield file_path


def compute_global_palette_by_histogram(pillow_imgs, bits_per_channel=5, palette_size=256):
    """Computes a global color palette for a sequence of images using histogram analysis.

    Args:
        pillow_imgs (list): List of pillow images. All images should be in RGB mode.
        bits_per_channel (int, optional): Number of bits to use per color channel for histogram
            binning. Can take values between 1 and 7. Defaults to 5.
        palette_size (int, optional): Number of colors in the resulting palette. Defaults to 256.

    Returns:
        PIL.Image: A PIL 'P' mode image containing the computed color palette.

    Raises:
        ValueError: If bits_per_channel or palette_size is outside of range.
    """

    if not 1 <= bits_per_channel <= 7:
        raise ValueError(f"bits_per_channel must be between 1 and 7, got {bits_per_channel}")
    if not 1 <= palette_size <= 256:
        raise ValueError(f"palette_size must be between 1 and 256, got {palette_size}")

    # compute number of bins per channel by bitshift
    bins_per = 1 << bits_per_channel
    # compute total number of histogram bins for RGB
    total_bins = bins_per**3
    # counts per bin in the final histogram
    counts = np.zeros(total_bins, dtype=np.int64)

    shift = 8 - bits_per_channel
    # Iterate images, accumulate bin counts
    for img in pillow_imgs:
        arr = np.array(img.convert("RGB"), dtype=np.uint8).reshape(-1, 3)
        # reduce bits, compute bin index
        r = (arr[:, 0] >> shift).astype(np.int32)
        g = (arr[:, 1] >> shift).astype(np.int32)
        b = (arr[:, 2] >> shift).astype(np.int32)
        idx = (r * bins_per + g) * bins_per + b
        # accumulate counts
        bincount = np.bincount(idx, minlength=total_bins)
        counts += bincount

    # pick top bins
    top_idx = np.argpartition(-counts, palette_size - 1)[:palette_size]

    # sort top bins by frequency
    top_idx = top_idx[np.argsort(-counts[top_idx])]

    # convert bin index back to representative RGB (center of bin)
    bins = np.array(
        [((i // (bins_per * bins_per)), (i // bins_per) % bins_per, i % bins_per) for i in top_idx]
    )

    # expand bin centers back to 8-bit values
    center = (
        (bins * (1 << (8 - bits_per_channel)) + (1 << (7 - bits_per_channel))).clip(0, 255)
    ).astype(np.uint8)
    palette_colors = center.reshape(-1, 3)  # shape (k,3)

    # build a PIL 'P' palette image from these colors
    pal = np.zeros(768, dtype=np.uint8)  # 256*3 entries
    pal[: palette_colors.size] = palette_colors.flatten()
    palette_img = Image.new("P", (1, 1))
    palette_img.putpalette(pal.tolist())

    return palette_img


def matplotlib_figure_to_numpy(fig, **kwargs):
    """Convert matplotlib figure to numpy array.

    Args:
        fig (matplotlib.figure.Figure): figure to convert.

    Returns:
        np.ndarray: numpy array of figure.

    """
    buf = BytesIO()
    fig.savefig(buf, format="png", bbox_inches="tight", **kwargs)
    buf.seek(0)
    image = Image.open(buf).convert("RGB")
    image = np.array(image)[..., :3]
    buf.close()
    return image


def retry_on_io_error(max_retries=3, initial_delay=0.5, retry_action=None):
    """Decorator to retry functions on I/O errors with exponential backoff.

    Args:
        max_retries (int): Maximum number of retry attempts.
        initial_delay (float): Initial delay between retries in seconds.
        retry_action (callable, optional): Optional function to call before each retry attempt.
            If decorating a method: ``retry_action(self, exception, attempt, *args, **kwargs)``
            If decorating a function: ``retry_action(exception, attempt, *args, **kwargs)``

    Returns:
        callable: Decorated function with retry logic.

    """

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            delay = initial_delay
            last_exception = None

            for attempt in range(max_retries):
                try:
                    return func(*args, **kwargs)
                except (OSError, IOError) as e:
                    last_exception = e

                    # if args exist and first arg is a class, update retry count of that method
                    if args and hasattr(args[0], "retry_count"):
                        args[0].retry_count = attempt + 1

                    if attempt < max_retries - 1:
                        # Execute custom retry action if provided
                        if retry_action:
                            # Pass all original arguments to retry_action
                            retry_action(
                                *args,
                                exception=e,
                                retry_count=attempt,
                                **kwargs,
                            )

                        time.sleep(delay)

                    else:
                        # Last attempt failed
                        log.error(f"Failed after {max_retries} attempts: {e}")

            # If we've exhausted all retries
            raise ValueError(
                f"Failed to complete operation after {max_retries} attempts. "
                f"Last error: {last_exception}"
            )

        return wrapper

    return decorator


def _convert_image_mode(img, mode="L"):
    """Convert a PIL Image to the specified mode and return as numpy array."""
    if mode not in {"L", "RGB"}:
        raise ValueError(f"Unsupported mode: {mode}, must be one of: L, RGB")
    if mode == "L":
        img = img.convert("L")
        arr = np.array(img)
    elif mode == "RGB":
        img = img.convert("RGB")
        arr = np.array(img)
    return arr


def grayscale_to_rgb(image):
    """Converts a grayscale image to an RGB image.

    Args:
        image (ndarray): Grayscale image. Must have shape (height, width).

    Returns:
        ndarray: RGB image.
    """
    assert image.ndim == 2, "Input image must be grayscale."
    # Stack the grayscale image into 3 channels (RGB)
    return np.stack([image] * 3, axis=-1)


def _assert_uint8_images(images: np.ndarray):
    """
    Asserts that the input images have the correct properties.

    Args:
        images (np.ndarray): The input images.

    Raises:
        AssertionError: If the dtype of images is not uint8.
        AssertionError: If the shape of images is not (n_frames, height, width, channels)
            or (n_frames, height, width) for grayscale images.
        AssertionError: If images have anything other than 1 (grayscale),
            3 (rgb) or 4 (rgba) channels.
    """
    assert images.dtype == np.uint8, f"dtype of images should be uint8, got {images.dtype}"

    assert images.ndim in (3, 4), (
        "images must have shape (n_frames, height, width, channels),"
        f" or (n_frames, height, width) for grayscale images. Got {images.shape}"
    )

    if images.ndim == 4:
        assert images.shape[-1] in (1, 3, 4), (
            "Grayscale images must have 1 channel, "
            "RGB images must have 3 channels, and RGBA images must have 4 channels. "
            f"Got shape: {images.shape}, channels: {images.shape[-1]}"
        )


def preprocess_for_saving(images):
    """Preprocesses images for saving to GIF or MP4.

    Args:
        images (ndarray, list[ndarray]): Images. Must have shape (n_frames, height, width, channels)
            or (n_frames, height, width).
    """
    images = np.array(images)

    if len(images) == 0:
        raise ValueError("Received empty list of images. Cannot save an empty video.")

    _assert_uint8_images(images)

    # Remove channel axis if it is 1 (grayscale image)
    if images.ndim == 4 and images.shape[-1] == 1:
        images = np.squeeze(images, axis=-1)

    # convert grayscale images to RGB
    if images.ndim == 3:
        images = [grayscale_to_rgb(image) for image in images]
        images = np.array(images)

    # drop alpha channel if present (RGBA -> RGB)
    if images.ndim == 4 and images.shape[-1] == 4:
        images = images[..., :3]

    return images



"""Command-line interface for copying a zea.Folder to a new location.

Usage:
    python -m zea.data <source_folder> <destination_folder> <key>
"""

import argparse

from zea import Folder


def get_parser():
    parser = argparse.ArgumentParser(description="Copy a :class:`zea.Folder` to a new location.")
    parser.add_argument("src", help="Source folder path")
    parser.add_argument("dst", help="Destination folder path")
    parser.add_argument("key", help="Key to access in the hdf5 files")
    parser.add_argument(
        "--mode",
        default="a",
        choices=["a", "w", "r+", "x"],
        help="Mode in which to open the destination files (default: 'a')",
    )
    return parser


def main():
    args = get_parser().parse_args()

    src_folder = Folder(args.src, validate=False)
    src_folder.copy(args.dst, args.key, mode=args.mode)


if __name__ == "__main__":
    main()



"""
Script to convert the EchoNet-LVH database to zea format.

Each video is cropped so that the scan cone is centered
without padding, such that it can be converted to polar domain.

.. note::
    This cropping requires first computing scan cone parameters
    using :mod:`zea.data.convert.echonetlvh.precompute_crop`, which
    are then passed to this script.

For more information about the dataset, resort to the following links:

- The original dataset can be found at `this link <https://stanfordaimi.azurewebsites.net/datasets/5b7fcc28-579c-4285-8b72-e4238eac7bd1>`_.
"""

import csv
import os
from concurrent.futures import ProcessPoolExecutor, as_completed
from pathlib import Path

import jax.numpy as jnp
import numpy as np
from jax import jit, vmap
from tqdm import tqdm

from zea import log
from zea.data import generate_zea_dataset
from zea.data.convert.echonet import H5Processor
from zea.data.convert.echonetlvh.precompute_crop import precompute_cone_parameters
from zea.data.convert.utils import load_avi, unzip
from zea.display import cartesian_to_polar_matrix
from zea.func.tensor import translate


def overwrite_splits(source_dir, rejection_path=None):
    """
    Overwrite MeasurementsList.csv splits based on manual_rejections.txt or another
    txt file specifying which hashes to reject.

    Args:
        source_dir: Source directory containing MeasurementsList.csv and manual_rejections.txt
        rejection_path: Path to the rejection txt file. If None, defaults to ./manual_rejections.txt
    Returns:
        None
    """
    current_dir = os.path.dirname(os.path.abspath(__file__))
    if rejection_path is None:
        rejection_path = os.path.join(current_dir, "manual_rejections.txt")
        expected_num_rejections = 278
    else:
        # unknown number of rejections for custom rejection file.
        # NOTE: this is used for testing, where we want to use a dummy rejections file
        expected_num_rejections = -1
    try:
        with open(rejection_path) as f:
            rejected_hashes = [line.strip() for line in f]
    except FileNotFoundError:
        log.warning(f"{rejection_path} not found, skipping rejections.")
        return

    csv_path = Path(source_dir) / "MeasurementsList.csv"
    temp_path = Path(source_dir) / "MeasurementsList_temp.csv"
    try:
        rejection_counter = 0
        with (
            csv_path.open("r", newline="", encoding="utf-8") as infile,
            temp_path.open("w", encoding="utf-8", newline="") as outfile,
        ):
            reader = csv.DictReader(infile)
            writer = csv.DictWriter(outfile, fieldnames=reader.fieldnames)
            writer.writeheader()
            for row in reader:
                if row["HashedFileName"] in rejected_hashes:
                    row["split"] = "rejected"
                    rejection_counter += 1
                writer.writerow(row)
            if expected_num_rejections != -1:
                assert rejection_counter == expected_num_rejections, (
                    f"Expected {expected_num_rejections} rejections, but applied only {rejection_counter}."
                )
    except FileNotFoundError:
        log.warning(f"{csv_path} not found, skipping rejections.")
        return
    temp_path.replace(csv_path)
    log.info(f"Overwritten {rejection_counter}/278 rejections to {csv_path}")
    return


def load_splits(source_dir):
    """
    Load splits from MeasurementsList.csv and return avi filenames

    Args:
        source_dir: Source directory containing MeasurementsList.csv
    Returns:
        Dictionary with keys 'train', 'val', 'test', 'rejected' and values as lists of avi filenames
    """
    csv_path = Path(source_dir) / "MeasurementsList.csv"
    splits = {"train": [], "val": [], "test": [], "rejected": []}
    with open(csv_path, newline="", encoding="utf-8") as csvfile:
        reader = csv.DictReader(csvfile)
        file_split_map = {}
        for row in reader:
            filename = row["HashedFileName"]
            split = row["split"]
            file_split_map.setdefault(filename, split)
        for filename, split in file_split_map.items():
            splits[split].append(filename + ".avi")
    return splits


def find_avi_file(source_dir, hashed_filename, batch=None):
    """
    Find AVI file in the specified batch directory or any batch if not specified.

    Args:
        source_dir: Source directory containing BatchX subdirectories
        hashed_filename: Hashed filename (with or without .avi extension)
        batch: Specific batch directory to search in (e.g., "Batch2"), or None to search all batches

    Returns:
        Path to the AVI file if found, else None
    """
    # If filename already has .avi extension, strip it
    if hashed_filename.endswith(".avi"):
        hashed_filename = hashed_filename[:-4]

    if batch:
        batch_dir = Path(source_dir) / batch
        avi_path = batch_dir / f"{hashed_filename}.avi"
        if avi_path.exists():
            return avi_path
        return None
    else:
        for batch_dir in Path(source_dir).glob("Batch*"):
            avi_path = batch_dir / f"{hashed_filename}.avi"
            if avi_path.exists():
                return avi_path
        return None


def load_cone_parameters(csv_path):
    """
    Load cone parameters from CSV file into a dictionary.

    Args:
        csv_path: Path to the CSV file containing cone parameters

    Returns:
        Dictionary mapping avi_filename to cone parameters
    """
    cone_params = {}

    with open(csv_path, "r", encoding="utf-8") as csvfile:
        reader = csv.DictReader(csvfile)
        for row in reader:
            if row["status"] == "success":
                # Convert string values to appropriate types
                params = {}
                for key, value in row.items():
                    if key in ("avi_filename", "status"):
                        params[key] = value
                    elif key == "apex_above_image":
                        params[key] = value.lower() == "true"
                    elif value is not None and value != "":
                        params[key] = float(value)
                    else:
                        params[key] = None

                cone_params[row["avi_filename"]] = params

    return cone_params


def crop_frame_with_params(frame, cone_params):
    """
    Crop a single frame using predetermined cone parameters.

    Args:
        frame: Input frame as numpy array
        cone_params: Dictionary containing cropping parameters

    Returns:
        Cropped and padded frame
    """
    crop_left = int(cone_params["crop_left"])
    crop_right = int(cone_params["crop_right"])
    crop_top = int(cone_params["crop_top"])
    crop_bottom = int(cone_params["crop_bottom"])

    # Handle negative crop_top
    if crop_top < 0:
        cropped = frame[0:crop_bottom, crop_left:crop_right]
        # Add top padding
        top_padding = -crop_top
        top_pad = jnp.zeros((top_padding, cropped.shape[1]), dtype=cropped.dtype)
        cropped = jnp.concatenate([top_pad, cropped], axis=0)
    else:
        cropped = frame[crop_top:crop_bottom, crop_left:crop_right]

    # Apply horizontal centering
    apex_x_in_crop = cone_params["apex_x"] - crop_left
    cropped_height, cropped_width = cropped.shape
    target_center_x = cropped_width / 2
    left_padding_needed = target_center_x - apex_x_in_crop

    left_padding = max(0, int(left_padding_needed))
    right_padding = max(0, int(-left_padding_needed))

    if left_padding > 0 or right_padding > 0:
        if left_padding > 0:
            left_pad = jnp.zeros((cropped_height, left_padding), dtype=cropped.dtype)
            cropped = jnp.concatenate([left_pad, cropped], axis=1)

        if right_padding > 0:
            right_pad = jnp.zeros((cropped_height, right_padding), dtype=cropped.dtype)
            cropped = jnp.concatenate([cropped, right_pad], axis=1)

    return cropped


def crop_sequence_with_params(sequence, cone_params):
    """
    Apply cropping to a sequence of frames using predetermined parameters.

    Args:
        sequence: Input sequence as numpy array of shape (frames, height, width)
        cone_params: Dictionary containing cropping parameters

    Returns:
        Cropped and padded sequence
    """
    crop_sequence = vmap(lambda frame: crop_frame_with_params(frame, cone_params))
    return crop_sequence(sequence)


class LVHProcessor(H5Processor):
    """Modified H5Processor for EchoNet-LVH dataset."""

    def __init__(self, *args, cone_params=None, **kwargs):
        super().__init__(*args, **kwargs)
        # Store the pre-computed cone parameters
        self.cart2pol_jit = jit(cartesian_to_polar_matrix)
        self.cart2pol_batched = vmap(
            (lambda matrix, angle: self.cart2pol_jit(matrix, angle=angle)), in_axes=(0, None)
        )  # map over sequence of images, keep the angle fixed since it's constant across a sequence
        self.cone_parameters = cone_params or {}

    def get_split(self, avi_file: str, sequence):
        """
        Get the split (train/val/test) for a given AVI file.

        Args:
            avi_file: Path to the AVI file
            sequence: Video sequence (unused)

        Returns:
            String indicating the split ('train', 'val', or 'test')
        """
        # Extract base filename without extension
        filename = Path(avi_file).stem + ".avi"

        for split, files in self.splits.items():
            if filename in files:
                return split
        raise UserWarning("Unknown split for file: " + filename)

    def __call__(self, avi_file):
        """Takes a single avi_file and generates a zea dataset

        Args:
            avi_file: String or path to avi_file to be processed

        Returns:
            zea dataset
        """

        avi_filename = Path(avi_file).stem + ".avi"
        sequence_np = load_avi(avi_file)
        sequence_processed = jnp.array(sequence_np)
        sequence_processed = translate(sequence_processed, self.range_from, self._process_range)
        # Get pre-computed cone parameters for this file
        cone_params = self.cone_parameters.get(avi_filename)
        if cone_params is not None:
            # Apply pre-computed cropping parameters
            sequence_processed = crop_sequence_with_params(sequence_processed, cone_params)
        else:
            raise UserWarning(f"No cone parameters for {avi_filename}")

        split = self.get_split(avi_file, sequence_processed)
        out_h5 = self.path_out_h5 / split / (Path(avi_file).stem + ".hdf5")

        angle = cone_params["opening_angle"] / 2  # angular field spans (-angle, +angle)
        polar_im_set = self.cart2pol_batched(sequence_processed, angle)
        sequence_processed = translate(sequence_processed, self._process_range, self.range_from)
        sequence_processed_uint8 = jnp.asarray(jnp.floor(sequence_processed + 0.5), dtype=jnp.uint8)
        del sequence_processed

        polar_im_set = translate(polar_im_set, self._process_range, (0, 255))
        polar_im_set_uint8 = jnp.asarray(jnp.floor(polar_im_set + 0.5), dtype=jnp.uint8)
        del polar_im_set

        if jnp.all(sequence_processed_uint8 == 0):
            raise ValueError(f"Processed sequence is all zeros for file {avi_file}")

        if jnp.all(polar_im_set_uint8 == 0):
            raise ValueError(f"Polar sequence is all zeros for file {avi_file}")

        zea_dataset = {
            "path": out_h5,
            "image_sc": sequence_processed_uint8,
            "probe_name": "generic",
            "description": "EchoNet-LVH dataset converted to zea format",
            "image": polar_im_set_uint8,
            "cast_to_float": False,
        }
        return generate_zea_dataset(**zea_dataset)


def transform_measurement_coordinates_with_cone_params(row, cone_params):
    """Transform measurement coordinates using cone parameters from fit_scan_cone.

    Args:
        row: A dict containing measurement data with X1,X2,Y1,Y2 coordinates
        cone_params: Dictionary containing cone parameters from fit_scan_cone

    Returns:
        A new row with transformed coordinates, or None if cone_params is None
    """
    if cone_params is None:
        log.warning(f"No cone parameters for file {row['HashedFileName']}")
        return None

    new_row = dict(row)

    # Apply cropping offset
    crop_left = cone_params["crop_left"]
    crop_top = cone_params["crop_top"]

    # Transform coordinates
    for k in ["X1", "X2", "Y1", "Y2"]:
        # Convert to float if not already
        new_row[k] = float(row[k]) - (crop_left if k.startswith("X") else crop_top)

    # Apply horizontal centering offset
    apex_x_in_crop = cone_params["apex_x"] - crop_left
    original_width = cone_params["crop_right"] - cone_params["crop_left"]
    target_center_x = original_width / 2
    left_padding_needed = target_center_x - apex_x_in_crop
    left_padding = max(0, int(left_padding_needed))

    # Adjust x coordinates for horizontal padding
    new_row["X1"] = new_row["X1"] + left_padding
    new_row["X2"] = new_row["X2"] + left_padding

    # Check if coordinates are within the final image bounds
    final_width = cone_params["new_width"]
    final_height = cone_params["new_height"]

    # Check if coordinates are out of bounds
    is_out_of_bounds = (
        new_row["X1"] < 0
        or new_row["X2"] < 0
        or new_row["Y1"] < 0
        or new_row["Y2"] < 0
        or new_row["X1"] >= final_width
        or new_row["X2"] >= final_width
        or new_row["Y1"] >= final_height
        or new_row["Y2"] >= final_height
    )

    if is_out_of_bounds:
        log.warning(f"Transformed coordinates out of bounds for file {row['HashedFileName']}")

    # Convert back to string if original was string
    for k in ["X1", "X2", "Y1", "Y2"]:
        new_row[k] = str(new_row[k])

    return new_row


def convert_measurements_csv(source_csv, output_csv, cone_params_csv=None):
    """Convert measurements CSV file with updated coordinates using cone parameters.

    Args:
        source_csv: Path to source CSV file
        output_csv: Path to output CSV file
        cone_params_csv: Path to CSV file with cone parameters
    """
    try:
        # Read the CSV file
        with open(source_csv, newline="", encoding="utf-8") as csvfile:
            reader = csv.DictReader(csvfile)
            rows = list(reader)
            fieldnames = reader.fieldnames

        # Load cone parameters if available
        cone_parameters = {}
        if cone_params_csv and Path(cone_params_csv).exists():
            cone_parameters = load_cone_parameters(cone_params_csv)
        else:
            log.warning("No cone parameters file found. Measurements will not be transformed.")

        # Apply coordinate transformation and track skipped rows
        transformed_rows = []
        skipped_files = set()

        for row in rows:
            try:
                avi_filename = row["HashedFileName"] + ".avi"
                cone_params = cone_parameters.get(avi_filename, None)
                transformed_row = transform_measurement_coordinates_with_cone_params(
                    row, cone_params
                )
                if transformed_row is not None:
                    transformed_rows.append(transformed_row)
                else:
                    skipped_files.add(row["HashedFileName"])
            except Exception as e:
                log.error(f"Error processing row for file {row['HashedFileName']}: {str(e)}")
                skipped_files.add(row["HashedFileName"])

        # Save to new CSV file
        if transformed_rows:
            # Use keys from first row as fieldnames
            out_fieldnames = list(transformed_rows[0].keys())
            with open(output_csv, "w", newline="", encoding="utf-8") as csvfile:
                writer = csv.DictWriter(csvfile, fieldnames=out_fieldnames)
                writer.writeheader()
                writer.writerows(transformed_rows)
        else:
            # Write header only if no rows
            with open(output_csv, "w", newline="", encoding="utf-8") as csvfile:
                writer = csv.DictWriter(csvfile, fieldnames=fieldnames)
                writer.writeheader()

        # Print summary
        log.info("Conversion Summary:")
        log.info(f"Total rows processed: {len(rows)}")
        log.info(f"Rows successfully converted: {len(transformed_rows)}")
        log.info(f"Rows skipped: {len(rows) - len(transformed_rows)}")
        if skipped_files:
            log.info("Skipped files:")
            for filename in sorted(skipped_files):
                log.info(f"  - {filename}")
        log.info(f"Converted measurements saved to {output_csv}")

    except Exception as e:
        log.error(f"Error processing CSV file: {str(e)}")
        raise


def _process_file_worker(avi_file, dst, splits, cone_parameters, range_from, process_range):
    """
    Function for a hyperthreading worker to process a single file.

    Args:
        avi_file: Path to the AVI file to process
        dst: Destination directory for output
        splits: Dictionary of splits
        cone_parameters: Dictionary of cone parameters
        range_from: Range from value for processing
        process_range: Process range value for processing
    Returns:
        Result of processing the file
    """

    # create a fresh processor inside the worker process
    proc = LVHProcessor(path_out_h5=dst, splits=splits, cone_params=cone_parameters)
    # if LVHProcessor needs range_from/_process_range set, set them here
    proc.range_from = range_from
    proc._process_range = process_range
    return proc(avi_file)


def convert_echonetlvh(args):
    """
    Conversion script for the EchoNet-LVH dataset.
    Unzips, overwrites splits if needed, precomputes cone parameters,
    and converts images and/or measurements to zea format and saves dataset.
    Is called with argparse arguments through zea/zea/data/convert/__main__.py

    Args:
        args (argparse.Namespace): Command-line arguments
    """
    # Check if unzip is needed
    src = unzip(args.src, "echonetlvh")

    # Overwrite the splits if manual rejections are provided
    if not args.no_rejection:
        overwrite_splits(args.src, getattr(args, "rejection_path", None))

    # Check that cone parameters exist
    cone_params_csv = Path(args.dst) / "cone_parameters.csv"
    if not cone_params_csv.exists():
        precompute_cone_parameters(args)

    # If no specific conversion is requested, convert both
    if not (args.convert_measurements or args.convert_images):
        args.convert_measurements = True
        args.convert_images = True

    # Convert images if requested
    if args.convert_images:
        source_path = Path(src)
        splits = load_splits(source_path)

        # Load precomputed cone parameters
        cone_parameters = load_cone_parameters(cone_params_csv)
        log.info(f"Loaded cone parameters for {len(cone_parameters)} files")

        files_to_process = []
        for split_files in splits.values():
            for avi_filename in split_files:
                # Strip .avi if present
                base_filename = avi_filename[:-4] if avi_filename.endswith(".avi") else avi_filename
                avi_file = find_avi_file(src, base_filename, batch=args.batch)
                if avi_file:
                    files_to_process.append(avi_file)
                else:
                    log.warning(
                        f"Warning: Could not find AVI file for {base_filename} in batch "
                        f"{args.batch if args.batch else 'any'}"
                    )

        # List files that have already been processed
        files_done = []
        for _, _, filenames in os.walk(args.dst):
            for filename in filenames:
                if filename.endswith(".hdf5"):
                    files_done.append(filename.replace(".hdf5", ""))

        # Filter out already processed files
        files_to_process = [f for f in files_to_process if f.stem not in files_done]

        # Limit files if max_files is specified
        if args.max_files is not None:
            files_to_process = files_to_process[: args.max_files]
            log.info(f"Limited to processing {args.max_files} files due to max_files parameter")

        log.info(f"Files left to process: {len(files_to_process)}")

        # Initialize processor with splits and cone parameters
        processor = LVHProcessor(path_out_h5=args.dst, splits=splits, cone_params=cone_parameters)

        log.info("Starting the conversion process.")

        for file in tqdm(files_to_process):
            try:
                processor(file)
            except Exception as e:
                log.error(f"Error processing {file}: {str(e)}")

        log.info("All image conversion tasks are completed.")

    # Convert measurements if requested
    if args.convert_measurements:
        source_path = Path(src)
        measurements_csv = source_path / "MeasurementsList.csv"
        if measurements_csv.exists():
            output_csv = Path(args.dst) / "MeasurementsList.csv"
            convert_measurements_csv(measurements_csv, output_csv, cone_params_csv)
        else:
            log.warning("MeasurementsList.csv not found in source directory")

    log.info("All tasks are completed.")
